from __future__ import annotations

import operator
import enum
from typing import TYPE_CHECKING, Any, Optional, Tuple, Union

from ._dtypes import (
    _boolean_dtypes,
    _integer_dtypes,
    _integer_or_boolean_dtypes,
    _floating_dtypes,
    _numeric_dtypes,
    _result_type,
    _dtype_categories,
)

if TYPE_CHECKING:
    from ._typing import cpu, Device, Dtype, PyCapsule

    try:
        from ._typing import gpu
    except ImportError:
        pass
    from builtins import ellipsis

import heat as ht
from heat import array_api


class Array:
    """
    DNDarray object for the array API namespace.
    This is a wrapper around ``heat.DNDarray`` that restricts the usage to only
    those things that are required by the array API namespace. Note,
    attributes on this object that start with a single underscore are not part
    of the API specification and should only be used internally. This object
    should not be constructed directly. Rather, use one of the creation
    functions, such as ``asarray``.
    """

    _array: ht.DNDarray

    @classmethod
    def _new(cls, x, /):
        """
        Initializes the array API Array
        object.
        Functions outside of the array_api submodule should not use this
        method. Use one of the creation functions instead, such as
        ``asarray``.

        Parameters
        ----------
        x : DNDarray
            Underlying ``DNDarray``
        """
        obj = super().__new__(cls)
        obj._array = x
        return obj

    def __new__(cls, *args, **kwargs):
        """
        Prevent ``Array()`` from working.
        """
        raise TypeError(
            "The array_api Array object should not be instantiated directly. Use an array creation function, such as asarray(), instead."
        )

    def __str__(self: Array, /) -> str:
        """
        Computes a printable representation of the Array.
        """
        return self._array.__str__().replace("DNDarray", "Array")

    def __repr__(self: Array, /) -> str:
        """
        Computes a printable representation of the Array.
        """
        return self._array.__str__().replace("DNDarray", "Array")

    def __len__(self) -> int:
        """
        The length of the Array.
        """
        return self._array.__len__()

    def _check_allowed_dtypes(
        self, other: Union[bool, int, float, Array], dtype_category: str, op: str
    ) -> Array:
        """
        Helper function for operators to only allow specific input dtypes
        Use like
            other = self._check_allowed_dtypes(other, 'numeric', '__add__')
            if other is NotImplemented:
                return other
        """
        if self.dtype not in _dtype_categories[dtype_category]:
            raise TypeError(f"Only {dtype_category} dtypes are allowed in {op}")
        if isinstance(other, (int, float, bool)):
            other = self._promote_scalar(other)
        elif isinstance(other, Array):
            if other.dtype not in _dtype_categories[dtype_category]:
                raise TypeError(f"Only {dtype_category} dtypes are allowed in {op}")
        else:
            return NotImplemented
        _result_type(self.dtype, other.dtype)
        return other

    def _promote_scalar(self, scalar):
        """
        Returns a promoted version of a Python scalar appropriate for use with
        operations on self.
        This may raise a ``TypeError`` when the scalar type is incompatible with
        the dtype of self.
        """
        # Note: Only Python scalar types that match the array dtype are
        # allowed.
        if isinstance(scalar, bool):
            if self.dtype not in _boolean_dtypes:
                raise TypeError("Python bool scalars can only be promoted with bool arrays")
        elif isinstance(scalar, int):
            if self.dtype in _boolean_dtypes:
                raise TypeError("Python int scalars cannot be promoted with bool arrays")
        elif isinstance(scalar, float):
            if self.dtype not in _floating_dtypes:
                raise TypeError(
                    "Python float scalars can only be promoted with floating-point arrays."
                )
        else:
            raise TypeError(f"{scalar} must be a Python scalar")

        return Array._new(ht.array(scalar, self.dtype))

    @staticmethod
    def _normalize_two_args(x1, x2) -> Tuple[Array, Array]:
        """
        Normalize inputs to two arg functions to fix type promotion rules
        Heat deviates from the spec type promotion rules in cases where one
        argument is 0-dimensional and the other is not. For example:
        >>> import heat as ht
        >>> a = ht.array([1.0], dtype=ht.float32)
        >>> b = ht.array(1.0, dtype=ht.float64)
        >>> ht.add(a, b)  # The spec says this should be float64
        DNDarray([2.], dtype=ht.float32, device=cpu:0, split=None)
        To fix this, we add a dimension to the 0-dimension array before passing it
        through. This works because a dimension would be added anyway from
        broadcasting, so the resulting shape is the same, but this prevents Heat
        from not promoting the dtype.
        """
        if x1.ndim == 0 and x2.ndim != 0:
            x1 = Array._new(x1._array[None])
        elif x2.ndim == 0 and x1.ndim != 0:
            x2 = Array._new(x2._array[None])
        return (x1, x2)

    def _validate_index(self, key):
        """
        Validate an index according to the array API.
        The array API specification only requires a subset of indices that are
        supported by Heat. This function will reject any index that is
        allowed by Heat but not required by the array API specification.
        This function raises IndexError if the index ``key`` is invalid.
        """
        _key = key if isinstance(key, tuple) else (key,)
        for i in _key:
            if isinstance(i, bool) or not (
                isinstance(i, int)  # i.e. ints
                or isinstance(i, slice)
                or i == Ellipsis
                or i is None
                or isinstance(i, Array)
                or isinstance(i, ht.DNDarray)
            ):
                raise IndexError(
                    f"Single-axes index {i} has {type(i)=}, but only "
                    "integers, slices (:), ellipsis (...), newaxis (None), "
                    "zero-dimensional integer arrays and boolean arrays "
                    "are specified in the Array API."
                )

        nonexpanding_key = []
        single_axes = []
        n_ellipsis = 0
        key_has_mask = False
        for i in _key:
            if i is not None:
                nonexpanding_key.append(i)
                if isinstance(i, Array) or isinstance(i, ht.DNDarray):
                    if i.dtype in _boolean_dtypes:
                        key_has_mask = True
                    single_axes.append(i)
                else:
                    # i must not be an array here, to avoid elementwise equals
                    if i == Ellipsis:
                        n_ellipsis += 1
                    else:
                        single_axes.append(i)

        n_single_axes = len(single_axes)
        if n_ellipsis > 1:
            return  # handled by DNDarray
        elif n_ellipsis == 0:
            # Note boolean masks must be the sole index, which we check for
            # later on.
            if not key_has_mask and n_single_axes < self.ndim:
                raise IndexError(
                    f"{self.ndim=}, but the multi-axes index only specifies "
                    f"{n_single_axes} dimensions. If this was intentional, "
                    "add a trailing ellipsis (...) which expands into as many "
                    "slices (:) as necessary."
                )

        if n_ellipsis == 0:
            indexed_shape = self.shape
        else:
            ellipsis_start = None
            for pos, i in enumerate(nonexpanding_key):
                if not (isinstance(i, Array) or isinstance(i, ht.DNDarray)):
                    if i == Ellipsis:
                        ellipsis_start = pos
                        break
            assert ellipsis_start is not None  # sanity check
            ellipsis_end = self.ndim - (n_single_axes - ellipsis_start)
            indexed_shape = self.shape[:ellipsis_start] + self.shape[ellipsis_end:]
        for i, side in zip(single_axes, indexed_shape):
            if isinstance(i, slice):
                if side == 0:
                    f_range = "0 (or None)"
                else:
                    f_range = f"between -{side} and {side - 1} (or None)"
                if i.start is not None:
                    try:
                        start = operator.index(i.start)
                    except TypeError:
                        raise IndexError("Invalid start value in slice")
                    else:
                        if not (-side <= start <= side):
                            raise IndexError(
                                f"Slice {i} contains {start=}, but should be "
                                f"{f_range} for an axis of size {side} "
                                "(out-of-bounds starts are not specified in "
                                "the Array API)"
                            )
                if i.stop is not None:
                    try:
                        stop = operator.index(i.stop)
                    except TypeError:
                        raise IndexError("Invalid stop value in slice")
                    else:
                        if not (-side <= stop <= side):
                            raise IndexError(
                                f"Slice {i} contains {stop=}, but should be "
                                f"{f_range} for an axis of size {side} "
                                "(out-of-bounds stops are not specified in "
                                "the Array API)"
                            )
            elif isinstance(i, Array):
                if i.dtype in _boolean_dtypes and len(_key) != 1:
                    assert isinstance(key, tuple)  # sanity check
                    raise IndexError(
                        f"Single-axes index {i} is a boolean array and "
                        f"{len(key)=}, but masking is only specified in the "
                        "Array API when the array is the sole index."
                    )
                elif i.dtype in _integer_dtypes and i.ndim != 0:
                    raise IndexError(
                        f"Single-axes index {i} is a non-zero-dimensional "
                        "integer array, but advanced integer indexing is not "
                        "specified in the Array API."
                    )
            elif isinstance(i, tuple):
                raise IndexError(
                    f"Single-axes index {i} is a tuple, but nested tuple "
                    "indices are not specified in the Array API."
                )

    def __abs__(self: Array, /) -> Array:
        """
        Calculates the absolute value for each element of an array instance
        (i.e., the element-wise result has the same magnitude as the respective
        element but has positive sign).
        """
        if self.dtype not in _numeric_dtypes:
            raise TypeError("Only numeric dtypes are allowed in __abs__")
        res = self._array.abs(dtype=self.dtype)
        return self.__class__._new(res)

    def __add__(self: Array, other: Union[int, float, Array], /) -> Array:
        """
        Calculates the sum for each element of an array instance with the
        respective element of the array ``other``.

        Parameters
        ----------
        other : Union[int, float, Array]
            Addend array. Must have a numeric data type.
        """
        other = self._check_allowed_dtypes(other, "numeric", "__add__")
        if other is NotImplemented:
            return other
        self, other = self._normalize_two_args(self, other)
        res = self._array.__add__(other._array)
        return self.__class__._new(res)

    def __and__(self: Array, other: Union[int, bool, Array], /) -> Array:
        """
        Evaluates ``self_i & other_i`` for each element of an array instance
        with the respective element of the array ``other``.

        Parameters
        ----------
        other : Union[int, bool, Array]
            Other array. Must have an integer or boolean data type.
        """
        other = self._check_allowed_dtypes(other, "integer or boolean", "__and__")
        if other is NotImplemented:
            return other
        self, other = self._normalize_two_args(self, other)
        res = self._array.__and__(other._array)
        return self.__class__._new(res)

    def __array_namespace__(self: Array, /, *, api_version: Optional[str] = None) -> Any:
        """
        Returns an object that has all the array API functions on it.

        Parameters
        ----------
        api_version : Optional[str]
            string representing the version of the array API specification to
            be returned, in ``'YYYY.MM'`` form. If it is ``None`` (default), it
            returns the namespace corresponding to latest version of the
            array API specification.
        """
        if api_version is not None and api_version != "2021.12":
            raise ValueError(f"Unrecognized array API version: {api_version}")
        return array_api

    def __bool__(self: Array, /) -> bool:
        """
        Converts a zero-dimensional boolean array to a Python ``bool`` object.
        """
        if self._array.ndim != 0:
            raise TypeError("bool is only allowed on arrays with 0 dimensions")
        if self.dtype not in _boolean_dtypes:
            raise ValueError("bool is only allowed on boolean arrays")
        res = self._array.__bool__()
        return res

    def __dlpack__(self: Array, /, *, stream: Optional[Union[int, Any]] = None) -> PyCapsule:
        """
        Exports the array for consumption by ``from_dlpack()`` as a DLPack capsule.

        Parameters
        ----------
        stream : Optional[Union[int, Any]]
            For CUDA and ROCm, a Python integer representing a pointer to a stream,
            on devices that support streams.
        """
        return self._array.__array.__dlpack__(stream=stream)

    def __dlpack_device__(self: Array, /) -> Tuple[enum.Enum, int]:
        """
        Returns device type and device ID in DLPack format. Meant for use
        within ``from_dlpack()``.
        """
        return self._array.__array.__dlpack_device__()

    def __eq__(self: Array, other: Union[int, float, bool, Array], /) -> Array:
        """
        Computes the truth value of ``self_i == other_i`` for each element of an
        array instance with the respective element of the array ``other``.

        Parameters
        ----------
        other : Union[int, float, bool, Array]
            Other array.
        """
        # Even though "all" dtypes are allowed, we still require them to be
        # promotable with each other.
        other = self._check_allowed_dtypes(other, "all", "__eq__")
        if other is NotImplemented:
            return other
        self, other = self._normalize_two_args(self, other)
        res = self._array.__eq__(other._array)
        return self.__class__._new(res)

    def __float__(self: Array, /) -> float:
        """
        Converts a zero-dimensional floating-point array to a Python ``float`` object.
        """
        if self._array.ndim != 0:
            raise TypeError("float is only allowed on arrays with 0 dimensions")
        if self.dtype not in _floating_dtypes:
            raise ValueError("float is only allowed on floating-point arrays")
        res = self._array.__float__()
        return res

    def __floordiv__(self: Array, other: Union[int, float, Array], /) -> Array:
        """
        Evaluates ``self_i // other_i`` for each element of an array instance
        with the respective element of the array ``other``.

        Parameters
        ----------
        other : Union[int, float, Array]
            Other array. Must have a numeric data type.
        """
        other = self._check_allowed_dtypes(other, "numeric", "__floordiv__")
        if other is NotImplemented:
            return other
        self, other = self._normalize_two_args(self, other)
        res = self._array.__floordiv__(other._array)
        return self.__class__._new(res)

    def __ge__(self: Array, other: Union[int, float, Array], /) -> Array:
        """
        Computes the truth value of ``self_i >= other_i`` for each element of
        an array instance with the respective element of the array ``other``.

        Parameters
        ----------
        other : Union[int, float, Array]
            Other array. Must have a numeric data type.
        """
        other = self._check_allowed_dtypes(other, "numeric", "__ge__")
        if other is NotImplemented:
            return other
        self, other = self._normalize_two_args(self, other)
        res = self._array.__ge__(other._array)
        return self.__class__._new(res)

    def __getitem__(
        self: Array,
        key: Union[int, slice, ellipsis, Tuple[Union[int, slice, ellipsis], ...], Array],
        /,
    ) -> Array:
        """
        Returns ``self[key]``.

        Parameters
        ----------
        key : Union[int, slice, ellipsis, Tuple[Union[int, slice, ellipsis], ...], Array]
            Index key
        """
        # Note: Only indices required by the spec are allowed.
        self._validate_index(key)
        if isinstance(key, Array):
            key = key._array
        res = self._array.__getitem__(key)
        return self._new(res)

    def __gt__(self: Array, other: Union[int, float, Array], /) -> Array:
        """
        Computes the truth value of ``self_i > other_i`` for each element
        of an array instance with the respective element of the array ``other``.

        Parameters
        ----------
        other : Union[int, float, Array]
            Other array. Must have a numeric data type.
        """
        other = self._check_allowed_dtypes(other, "numeric", "__gt__")
        if other is NotImplemented:
            return other
        self, other = self._normalize_two_args(self, other)
        res = self._array.__gt__(other._array)
        return self.__class__._new(res)

    def __index__(self: Array, /) -> int:
        """
        Converts a zero-dimensional integer array to a Python ``int`` object.
        """
        return self.__int__()

    def __int__(self: Array, /) -> int:
        """
        Converts a zero-dimensional integer array to a Python ``int`` object.
        """
        if self._array.ndim != 0:
            raise TypeError("int is only allowed on arrays with 0 dimensions")
        if self.dtype not in _integer_dtypes:
            raise ValueError("int is only allowed on integer arrays")
        res = self._array.__int__()
        return res

    def __invert__(self: Array, /) -> Array:
        """
        Evaluates ``~self_i`` for each element of an array instance.
        """
        if self.dtype not in _integer_or_boolean_dtypes:
            raise TypeError("Only integer or boolean dtypes are allowed in __invert__")
        res = self._array.__invert__()
        return self.__class__._new(res)

    def __le__(self: Array, other: Union[int, float, Array], /) -> Array:
        """
        Computes the truth value of ``self_i <= other_i`` for each element of an
        array instance with the respective element of the array ``other``.

        Parameters
        ----------
        other : Union[int, float, Array]
            Other array. Must have a numeric data type.
        """
        other = self._check_allowed_dtypes(other, "numeric", "__le__")
        if other is NotImplemented:
            return other
        self, other = self._normalize_two_args(self, other)
        res = self._array.__le__(other._array)
        return self.__class__._new(res)

    def __lshift__(self: Array, other: Union[int, Array], /) -> Array:
        """
        Evaluates ``self_i << other_i`` for each element of an array instance
        with the respective element of the array ``other``.

        Parameters
        ----------
        other : Union[int, Array]
            Other array. Must have an integer data type. Each element
            must be greater than or equal to ``0``.
        """
        other = self._check_allowed_dtypes(other, "integer", "__lshift__")
        if other is NotImplemented:
            return other
        self, other = self._normalize_two_args(self, other)
        res = self._array.__lshift__(other._array)
        return self.__class__._new(res)

    def __lt__(self: Array, other: Union[int, float, Array], /) -> Array:
        """
        Computes the truth value of ``self_i < other_i`` for each element
        of an array instance with the respective element of the array ``other``.

        Parameters
        ----------
        other : Union[int, Array]
            Other array. Must have a numeric data type.
        """
        other = self._check_allowed_dtypes(other, "numeric", "__lt__")
        if other is NotImplemented:
            return other
        self, other = self._normalize_two_args(self, other)
        res = self._array.__lt__(other._array)
        return self.__class__._new(res)

    def __matmul__(self: Array, other: Array, /) -> Array:
        """
        Computes the matrix product.

        Parameters
        ----------
        other : Array
            Other array. Must have a numeric data type and at least one dimension.
        """
        other = self._check_allowed_dtypes(other, "numeric", "__matmul__")
        if other is NotImplemented:
            return other
        res = self._array.__matmul__(other._array)
        return self.__class__._new(res)

    def __mod__(self: Array, other: Union[int, float, Array], /) -> Array:
        """
        Evaluates ``self_i % other_i`` for each element of an array instance
        with the respective element of the array ``other``.

        Parameters
        ----------
        other : Union[int, float, Array]
            Other array. Must have a numeric data type.
        """
        other = self._check_allowed_dtypes(other, "numeric", "__mod__")
        if other is NotImplemented:
            return other
        self, other = self._normalize_two_args(self, other)
        res = self._array.__mod__(other._array)
        return self.__class__._new(res)

    def __mul__(self: Array, other: Union[int, float, Array], /) -> Array:
        """
        Calculates the product for each element of an array instance with
        the respective element of the array ``other``.

        Parameters
        ----------
        other : Union[int, float, Array]
            Other array. Must have a numeric data type.
        """
        other = self._check_allowed_dtypes(other, "numeric", "__mul__")
        if other is NotImplemented:
            return other
        self, other = self._normalize_two_args(self, other)
        res = self._array.__mul__(other._array)
        return self.__class__._new(res)

    def __ne__(self: Array, other: Union[int, float, bool, Array], /) -> Array:
        """
        Computes the truth value of ``self_i != other_i`` for each element of
        an array instance with the respective element of the array ``other``.

        Parameters
        ----------
        other : Union[int, float, bool, Array]
            Other array.
        """
        other = self._check_allowed_dtypes(other, "all", "__ne__")
        if other is NotImplemented:
            return other
        self, other = self._normalize_two_args(self, other)
        res = self._array.__ne__(other._array)
        return self.__class__._new(res)

    def __neg__(self: Array, /) -> Array:
        """
        Evaluates ``-self_i`` for each element of an array instance.
        """
        if self.dtype not in _numeric_dtypes:
            raise TypeError("Only numeric dtypes are allowed in __neg__")
        res = self._array.__neg__()
        return self.__class__._new(res)

    def __or__(self: Array, other: Union[int, bool, Array], /) -> Array:
        """
        Evaluates ``self_i | other_i`` for each element of an array instance
        with the respective element of the array ``other``.

        Parameters
        ----------
        other : Union[int, bool, Array]
            Other array. Must have an integer or boolean data type.
        """
        other = self._check_allowed_dtypes(other, "integer or boolean", "__or__")
        if other is NotImplemented:
            return other
        self, other = self._normalize_two_args(self, other)
        res = self._array.__or__(other._array)
        return self.__class__._new(res)

    def __pos__(self: Array, /) -> Array:
        """
        Evaluates ``+self_i`` for each element of an array instance.
        """
        if self.dtype not in _numeric_dtypes:
            raise TypeError("Only numeric dtypes are allowed in __pos__")
        res = self._array.__pos__()
        return self.__class__._new(res)

    def __pow__(self: Array, other: Union[int, float, Array], /) -> Array:
        """
        Calculates an approximation of exponentiation by raising each element
        (the base) of an array instance to the power of ``other_i`` (the exponent),
        where ``other_i`` is the corresponding element of the array ``other``.

        Parameters
        ----------
        other : Union[int, float, Array]
            Other array. Must have a numeric data type.
        """
        other = self._check_allowed_dtypes(other, "numeric", "__pow__")
        if other is NotImplemented:
            return other
        self, other = self._normalize_two_args(self, other)
        res = self._array.__pow__(other._array)
        return self.__class__._new(res)

    def __rshift__(self: Array, other: Union[int, Array], /) -> Array:
        """
        Evaluates ``self_i >> other_i`` for each element of an array instance
        with the respective element of the array ``other``.

        Parameters
        ----------
        other : Union[int, Array]
            Other array. Must have an integer data type. Each element must be
            greater than or equal to ``0``.
        """
        other = self._check_allowed_dtypes(other, "integer", "__rshift__")
        if other is NotImplemented:
            return other
        self, other = self._normalize_two_args(self, other)
        res = self._array.__rshift__(other._array)
        return self.__class__._new(res)

    def __setitem__(
        self,
        key: Union[int, slice, ellipsis, Tuple[Union[int, slice, ellipsis], ...], Array],
        value: Union[int, float, bool, Array],
        /,
    ) -> None:
        """
        Sets ``self[key]`` to ``value``.
        """
        # Note: Only indices required by the spec are allowed.
        self._validate_index(key)
        if isinstance(key, Array):
            key = key._array
        if isinstance(value, Array):
            value = value._array
        self._array.__setitem__(key, value)

    def __sub__(self: Array, other: Union[int, float, Array], /) -> Array:
        """
        Calculates the difference for each element of an array instance
        with the respective element of the array ``other``.

        Parameters
        ----------
        other : Union[int, float, Array]
            Subtrahend array. Must have a numeric data type.
        """
        other = self._check_allowed_dtypes(other, "numeric", "__sub__")
        if other is NotImplemented:
            return other
        self, other = self._normalize_two_args(self, other)
        res = self._array.__sub__(other._array)
        return self.__class__._new(res)

    def __truediv__(self: Array, other: Union[int, float, Array], /) -> Array:
        """
        Evaluates ``self_i / other_i`` for each element of an array instance
        with the respective element of the array ``other``.

        Parameters
        ----------
        other : Union[int, float, Array]
            Subtrahend array. Must have a numeric data type.
        """
        other = self._check_allowed_dtypes(other, "numeric", "__truediv__")
        if other is NotImplemented:
            return other
        self, other = self._normalize_two_args(self, other)
        res = self._array.__truediv__(other._array)
        return self.__class__._new(res)

    def __xor__(self: Array, other: Union[int, bool, Array], /) -> Array:
        """
        Evaluates ``self_i ^ other_i`` for each element of an array instance
        with the respective element of the array ``other``.

        Parameters
        ----------
        other : Union[int, bool, Array]
            Subtrahend array. Must have an integer or boolean data type.
        """
        other = self._check_allowed_dtypes(other, "integer or boolean", "__xor__")
        if other is NotImplemented:
            return other
        self, other = self._normalize_two_args(self, other)
        res = self._array.__xor__(other._array)
        return self.__class__._new(res)

    def __radd__(self: Array, other: Union[int, float, Array], /) -> Array:
        """
        Reflected version of ``__add__``.

        Parameters
        ----------
        other : Union[int, float, Array]
            Addend array. Must have a numeric data type.
        """
        other = self._check_allowed_dtypes(other, "numeric", "__radd__")
        if other is NotImplemented:
            return other
        self, other = self._normalize_two_args(self, other)
        res = self._array.__radd__(other._array)
        return self.__class__._new(res)

    def __rfloordiv__(self: Array, other: Union[int, float, Array], /) -> Array:
        """
        Reflected version of ``__floordiv__``.

        Parameters
        ----------
        other : Union[int, float, Array]
            Other array. Must have a numeric data type.
        """
        other = self._check_allowed_dtypes(other, "numeric", "__rfloordiv__")
        if other is NotImplemented:
            return other
        self, other = self._normalize_two_args(self, other)
        res = self._array.__rfloordiv__(other._array)
        return self.__class__._new(res)

    def __rmod__(self: Array, other: Union[int, float, Array], /) -> Array:
        """
        Reflected version of ``__rmod__``.
        """
        other = self._check_allowed_dtypes(other, "numeric", "__rmod__")
        if other is NotImplemented:
            return other
        self, other = self._normalize_two_args(self, other)
        res = self._array.__rmod__(other._array)
        return self.__class__._new(res)

    def __rmul__(self: Array, other: Union[int, float, Array], /) -> Array:
        """
        Reflected version of ``__mul__``.

        Parameters
        ----------
        other : Union[int, float, Array]
            Other array. Must have a numeric data type.
        """
        other = self._check_allowed_dtypes(other, "numeric", "__rmul__")
        if other is NotImplemented:
            return other
        self, other = self._normalize_two_args(self, other)
        res = self._array.__rmul__(other._array)
        return self.__class__._new(res)

    def __rpow__(self: Array, other: Union[int, float, Array], /) -> Array:
        """
        Reflected version of ``__rpow__``.

        Parameters
        ----------
        other : Union[int, float, Array]
            Other array. Must have a numeric data type.
        """
        other = self._check_allowed_dtypes(other, "numeric", "__rpow__")
        if other is NotImplemented:
            return other
        self, other = self._normalize_two_args(self, other)
        res = self._array.__rpow__(other._array)
        return self.__class__._new(res)

    def __rsub__(self: Array, other: Union[int, float, Array], /) -> Array:
        """
        Reflected version of ``__sub__``.

        Parameters
        ----------
        other : Union[int, float, Array]
            Subtrahend array. Must have a numeric data type.
        """
        other = self._check_allowed_dtypes(other, "numeric", "__rsub__")
        if other is NotImplemented:
            return other
        self, other = self._normalize_two_args(self, other)
        res = self._array.__rsub__(other._array)
        return self.__class__._new(res)

    def __rtruediv__(self: Array, other: Union[float, Array], /) -> Array:
        """
        Reflected version of ``__truediv__``.

        Parameters
        ----------
        other : Union[int, float, Array]
            Subtrahend array. Must have a numeric data type.
        """
        other = self._check_allowed_dtypes(other, "floating-point", "__rtruediv__")
        if other is NotImplemented:
            return other
        self, other = self._normalize_two_args(self, other)
        res = self._array.__rtruediv__(other._array)
        return self.__class__._new(res)

    def to_device(
        self: Array, device: Device, /, stream: Optional[Union[int, Any]] = None
    ) -> Array:
        """
        Copy the array from the device on which it currently resides to the specified ``device``.

        Parameters
        ----------
        device : Device
            A ``Device`` object.
        stream : Optional[Union[int, Any]]
            Stream object to use during copy.
        """
        if stream is not None:
            raise ValueError("The stream argument to to_device() is not supported")
        if device == cpu:
            return self._array.cpu()
        elif device == gpu:
            return self._array.gpu()
        raise ValueError(f"Unsupported device {device!r}")

    @property
    def dtype(self) -> Dtype:
        """
        Data type of the array elements.
        """
        return self._array.dtype

    @property
    def device(self) -> Device:
        """
        Hardware device the array data resides on.
        """
        return self._array.device

    @property
    def mT(self) -> Array:
        """
        Transpose of a matrix (or a stack of matrices).
        """
        from .linalg import matrix_transpose

        if self.ndim < 2:
            raise ValueError("x.mT requires x to have at least 2 dimensions.")
        return matrix_transpose(self)

    @property
    def ndim(self) -> int:
        """
        Number of array dimensions (axes).
        """
        return self._array.ndim

    @property
    def shape(self) -> Tuple[Optional[int], ...]:
        """
        Array dimensions.
        """
        return self._array.shape

    @property
    def size(self) -> Optional[int]:
        """
        Number of elements in an array.
        """
        return self._array.size

    @property
    def T(self) -> Array:
        """
        Transpose of the array.
        """
        # Note: T only works on 2-dimensional arrays, as outlined in the specification:
        if self.ndim != 2:
            raise ValueError(
                "x.T requires x to have 2 dimensions. Use x.mT to transpose stacks of matrices and permute_dims() to permute dimensions."
            )
        return self.__class__._new(self._array.T)



import heat as ht

e = ht.e
"""IEEE 754 floating-point representation of Euler's constant."""
inf = ht.inf
"""IEEE 754 floating-point representation of (positive) infinity."""
nan = ht.nan
"""IEEE 754 floating-point representation of Not a Number (``NaN``)."""
newaxis = None
"""An alias for ``None`` which is useful for indexing arrays."""
pi = ht.pi
"""IEEE 754 floating-point representation of the mathematical constant :math:`\\pi`."""



from __future__ import annotations

from ._array_object import Array
from ._dtypes import _result_type, _numeric_dtypes

from typing import Optional, Tuple

import heat as ht


def argmax(x: Array, /, *, axis: Optional[int] = None, keepdims: bool = False) -> Array:
    """
    Returns the indices of the maximum values along a specified axis. When the
    maximum value occurs multiple times, only the indices corresponding to the
    first occurrence are returned.

    Parameters
    ----------
    x : Array
        Input array. Must have a numeric data type.
    axis : Optional[int]
        Axis along which to search. If ``None``, the function returns the index of
        the maximum value of the flattened array. Default: ``None``.
    keepdims : bool
        If ``True``, the reduced axes (dimensions) are included in the result as
        singleton dimensions. Otherwise, if ``False``, the reduced axes (dimensions)
        are not be included in the result. Default: ``False``.
    """
    if x.dtype not in _numeric_dtypes:
        raise TypeError("Only numeric dtypes are allowed in argmax")
    res = ht.argmax(x._array, axis=axis, keepdim=keepdims)
    return Array._new(res)


def argmin(x: Array, /, *, axis: Optional[int] = None, keepdims: bool = False) -> Array:
    """
    Returns the indices of the minimum values along a specified axis. When the
    minimum value occurs multiple times, only the indices corresponding to the
    first occurrence are returned.

    Parameters
    ----------
    x : Array
        Input array. Must have a numeric data type.
    axis : Optional[int]
        Axis along which to search. If ``None``, the function returns the index of
        the minimum value of the flattened array. Default: ``None``.
    keepdims : bool
        If ``True``, the reduced axes (dimensions) are included in the result as
        singleton dimensions. Otherwise, if ``False``, the reduced axes (dimensions)
        are not be included in the result. Default: ``False``.
    """
    if x.dtype not in _numeric_dtypes:
        raise TypeError("Only numeric dtypes are allowed in argmin")
    res = ht.argmin(x._array, axis=axis, keepdim=keepdims)
    return Array._new(res)


def nonzero(x: Array, /) -> Tuple[Array, ...]:
    """
    Returns the indices of the array elements which are non-zero.

    Parameters
    ----------
    x : Array
        Input array. Must have a positive rank.
    """
    # See PR #914 for overhaul
    return tuple(Array._new(i) for i in ht.nonzero(x._array))


def where(condition: Array, x1: Array, x2: Array, /) -> Array:
    """
    Returns elements chosen from ``x1`` or ``x2`` depending on ``condition``.

    Parameters
    ----------
    condition : Array
        When ``True``, yield ``x1_i``; otherwise, yield ``x2_i``. Must be
        compatible with ``x1`` and ``x2``.
    x1 : Array
        First input array. Must be compatible with ``condition`` and ``x2``.
    x2 : Array
        Second input array. Must be compatible with ``condition`` and ``x1``.
    """
    # Call result type here just to raise on disallowed type combinations
    _result_type(x1.dtype, x2.dtype)
    x1, x2 = Array._normalize_two_args(x1, x2)
    return Array._new(ht.where(condition._array, x1._array, x2._array))



"""Generalized operations for DCSR_matrix"""

import torch
import numpy as np

from heat.sparse.dcsx_matrix import DCSC_matrix, DCSR_matrix, __DCSX_matrix

from . import factories
from ..core.communication import MPI
from ..core import types

from typing import Callable, Optional, Dict

__all__ = []


def __binary_op_csx(
    operation: Callable,
    t1: __DCSX_matrix,
    t2: __DCSX_matrix,
    out: Optional[__DCSX_matrix] = None,
    orientation: str = "row",
    fn_kwargs: Optional[Dict] = {},
) -> __DCSX_matrix:
    """
    Generic wrapper for element-wise binary operations of two operands.
    Takes the operation function and the two operands involved in the operation as arguments.

    Parameters
    ----------
    operation : PyTorch function
        The operation to be performed. Function that performs operation elements-wise on the involved tensors,
        e.g. add values from other to self
    t1: __DCSX_matrix or scalar
        The first operand involved in the operation.
    t2: __DCSX_matrix or scalar
        The second operand involved in the operation.
    out: __DCSX_matrix, optional
        Output buffer in which the result is placed. If not provided, a freshly allocated matrix is returned.
    orientation: str, optional
        The orientation of the operation. Options: 'row' or 'col'
        Default: 'row'
    fn_kwargs: Dict, optional
        keyword arguments used for the given operation
        Default: {} (empty dictionary)

    Returns
    -------
    result: ht.sparse.__DCSX_matrix
        A __DCSX_matrix containing the results of element-wise operation.

    Raises
    ------
    ValueError
        If the orientation is invalid
    ValueError
        If the input types are not supported
    ValueError
        If the input shapes are not compatible
    ValueError
        If the output buffer shape is not compatible with the result
    """
    if orientation not in ["row", "col"]:
        raise ValueError(f"Invalid orientation: '{orientation}'. Options: 'row' or 'col'")

    if not np.isscalar(t1) and not isinstance(t1, __DCSX_matrix):
        raise TypeError(
            f"Only Dcsr_matrices and numeric scalars are supported, but input was {type(t1)}"
        )
    if not np.isscalar(t2) and not isinstance(t2, __DCSX_matrix):
        raise TypeError(
            f"Only Dcsr_matrices and numeric scalars are supported, but input was {type(t2)}"
        )

    if not isinstance(t1, __DCSX_matrix) and not isinstance(t2, __DCSX_matrix):
        raise TypeError(
            f"Operator only to be used with Dcsr_matrices, but input types were {type(t1)} and {type(t2)}"
        )

    promoted_type = types.result_type(t1, t2).torch_type()

    torch_constructor = torch.sparse_csr_tensor if orientation == "row" else torch.sparse_csc_tensor
    factory_method = (
        factories.sparse_csr_matrix if orientation == "row" else factories.sparse_csc_matrix
    )
    split_axis = 0 if orientation == "row" else 1

    # If one of the inputs is a scalar
    # just perform the operation on the data tensor
    # and create a new sparse matrix
    if np.isscalar(t1) or np.isscalar(t2):
        matrix = t1
        scalar = t2

        if np.isscalar(t1):
            matrix = t2
            scalar = t1

        res_values = operation(matrix.larray.values().to(promoted_type), scalar, **fn_kwargs)
        res_torch_sparse_csx = torch_constructor(
            matrix.lindptr,
            matrix.lindices,
            res_values,
            size=matrix.lshape,
            device=matrix.device.torch_device,
        )
        return factory_method(
            res_torch_sparse_csx, is_split=matrix.split, comm=matrix.comm, device=matrix.device
        )

    if t1.shape != t2.shape:
        raise ValueError(
            f"Dcsr_matrices of different shapes are not supported, but input shapes were {t1.shape} and {t2.shape}"
        )
    output_shape = t1.shape

    if t1.split is not None or t2.split is not None:
        if t1.split is None:
            t1 = factory_method(t1.larray, split=split_axis)

        if t2.split is None:
            t2 = factory_method(t2.larray, split=split_axis)

    output_split = t1.split
    output_device = t1.device
    output_comm = t1.comm
    output_balanced = t1.balanced
    output_lshape = t1.lshape

    # sanitize out buffer
    if out is not None:
        if out.shape != output_shape:
            raise ValueError(
                f"Output buffer shape is not compatible with the result. Expected {output_shape}, received {out.shape}"
            )

        if out.split != output_split:
            if out.split is None:
                out = factory_method(out.larray, split=split_axis)
            else:
                out = factory_method(
                    torch_constructor(
                        torch.tensor(out.indptr, dtype=torch.int64),
                        torch.tensor(out.indices, dtype=torch.int64),
                        torch.tensor(out.data),
                    )
                )

        out.device = output_device
        out.balanced = (
            output_balanced  # At this point, inputs and out buffer assumed to be balanced
        )

    # If there are no non-zero elements in either tensors, skip torch operation to
    #   1. Avoid unnecessary computation
    #   2. Avoid undefined behaviour when no data in process
    if t1.lnnz == 0 and t2.lnnz == 0:
        result = t1.larray
    else:
        result = operation(t1.larray.to(promoted_type), t2.larray.to(promoted_type), **fn_kwargs)

    if output_split is not None:
        output_gnnz = torch.tensor(result._nnz())
        output_comm.Allreduce(MPI.IN_PLACE, output_gnnz, MPI.SUM)
        output_gnnz = output_gnnz.item()
    else:
        output_gnnz = torch.tensor(result._nnz())

    output_type = types.canonical_heat_type(result.dtype)

    if out is None:
        if orientation == "row":
            return DCSR_matrix(
                array=torch_constructor(
                    result.crow_indices().to(torch.int64),
                    result.col_indices().to(torch.int64),
                    result.values(),
                    size=output_lshape,
                ),
                gnnz=output_gnnz,
                gshape=output_shape,
                dtype=output_type,
                split=output_split,
                device=output_device,
                comm=output_comm,
                balanced=output_balanced,
            )
        else:
            return DCSC_matrix(
                array=torch_constructor(
                    result.ccol_indices().to(torch.int64),
                    result.row_indices().to(torch.int64),
                    result.values(),
                    size=output_lshape,
                ),
                gnnz=output_gnnz,
                gshape=output_shape,
                dtype=output_type,
                split=output_split,
                device=output_device,
                comm=output_comm,
                balanced=output_balanced,
            )

    out.larray.copy_(result)
    out.gnnz = output_gnnz
    out.dtype = output_type
    out.comm = output_comm
    return out



"""
file for future "full" SVD implementation
"""

from typing import Tuple
from ..dndarray import DNDarray
from .qr import qr
from .polar import polar
from .eigh import eigh
from ..types import float32, float64
import torch
from warnings import warn

__all__ = ["svd"]


def svd(
    A: DNDarray,
    full_matrices: bool = False,
    compute_uv: bool = True,
    qr_procs_to_merge: int = 2,
    r_max_zolopd: int = 8,
) -> Tuple[DNDarray, DNDarray, DNDarray]:
    """
    Computes the singular value decomposition of a matrix (the input array ``A``).
    For an input DNDarray ``A`` of shape ``(M, N)``, the function returns DNDarrays ``U``, ``S``, and ``V.T`` such that ``A = U @ ht.diag(S) @ V.T``
    with shapes ``(M, min(M,N))``, ``(min(M, N),)``, and ``(min(M,N),N)``, respectively, in the case that ``compute_uv=True``, or
    only the vector containing the singular values ``S`` of shape ``(min(M, N),)`` in the case that ``compute_uv=False``. By definition of the singular value decomposition,
    the matrix ``U`` is orthogonal, the matrix ``V`` is orthogonal, and the entries of the vector ``S``are non-negative real numbers.

    We refer to, e.g., wikipedia (https://en.wikipedia.org/wiki/Singular_value_decomposition) or to Gene H. Golub and Charles F. Van Loan, Matrix Computations (3rd Ed., 1996),
    for more detailed information on the singular value decomposition.

    Parameters
    ----------
    A : ht.DNDarray
        The input array (2D, float32 or float64) for which the singular value decomposition is computed.
        Must be tall skinny (``M >> N``) or short fat (``M << n``) for the current implementation; an implementation that covers the remaining cases is planned.
    full_matrices : bool, optional
        currently, only the default value ``False`` is supported. This argument is included for compatibility with NumPy.
    compute_uv : bool, optional
        if ``True``, the matrices ``U`` and ``V.T`` are computed and returned together with the singular values ``S``.
        If ``False``, only the vector ``S`` containing the singular values is returned.
    qr_procs_to_merge : int, optional
        the number of processes to merge in the tall skinny QR decomposition that is applied if the input array is tall skinny (``M > N``) or short fat (``M < N``).
        See the corresponding remarks for :func:``heat.linalg.qr`` for more details.
    r_max_zolopd : int, optional
        an internal parameter only relevant for the case that the input matrix is neither tall-skinny nor short-fat.
        This parameter is passed to the Zolotarev-Polar Decomposition and the symmetric eigenvalue decomposition that is applied in this case.
        See the documentation of :func:``heat.linalg.polar`` as well as of :func:``heat.linalg.eigh`` for more details.

    Notes
    -----
    Unlike in NumPy, we currently do not support the option ``full_matrices=True``, since this can result in heavy memory consumption (in particular for tall skinny
    and short fat matrices) that should be avoided in the context Heat is designed for. If you nevertheless require this feature, please open an issue on GitHub.

    The algorithm used for the computation of the singular value depends on the shape of the input array ``A``.
    For tall and skinny matrices (``M > N``), the algorithm is based on the tall-skinny QR decomposition. For the remaining cases we use the approach based on
    Zolotarev-Polar Decomposition and a symmetric eigenvalue decomposition based on Zolotarev-Polar Decomposition; see Algorithm 5.3 in:

        Nakatsukasa, Y., & Freund, R. W. (2016). Computing fundamental matrix decompositions accurately via the
        matrix sign function in two iterations: The power of Zolotarev's functions. SIAM Review, 58(3).

    See Also
    --------
    :func:`heat.linalg.qr`
    :func:`heat.linalg.polar`
    :func:`heat.linalg.eigh`
    """
    if full_matrices:
        raise NotImplementedError(
            "SVD with 'full_matrices=True' is not supported by Heat yet as it might result in heavy memory usage. \n Please open an issue on GitHub if you nevertheless require this feature."
        )

    if not isinstance(A, DNDarray):
        raise TypeError(f"'A' must be a DNDarray, but is {type(A)}")
    if not isinstance(qr_procs_to_merge, int):
        raise TypeError(
            f"procs_to_merge must be an int, but is currently {type(qr_procs_to_merge)}"
        )
    if qr_procs_to_merge < 0 or qr_procs_to_merge == 1:
        raise ValueError(
            f"procs_to_merge must be 0 (for merging all processes at once) or at least 2, but is currently {qr_procs_to_merge}"
        )
    if qr_procs_to_merge == 0:
        qr_procs_to_merge = A.comm.size
    if not isinstance(r_max_zolopd, int) or r_max_zolopd < 0 or r_max_zolopd > 8:
        raise ValueError(
            f"r_max_zolopd must be a non-negative int, but is currently {r_max_zolopd} of type {type(r_max_zolopd)}"
        )
    if A.ndim != 2:
        raise ValueError(
            f"Array ``A`` must be 2 dimensional, buts has {A.ndim} dimensions. \n Please open an issue on GitHub if you require SVD for batches of matrices similar to PyTorch."
        )
    if A.dtype not in [float32, float64]:
        raise TypeError(
            f"Array ``A`` must have a datatype of float32 or float64, but has {A.dtype}"
        )

    def _toDNDarray(array):
        """Returns an unsplit heat DNDarray that inherits properties from the `A` matrix to be decomposed"""
        return DNDarray(
            array,
            tuple(array.shape),
            dtype=A.dtype,
            split=None,
            device=A.device,
            comm=A.comm,
            balanced=A.balanced,
        )

    if not A.is_distributed():
        # this is the non-distributed case
        if compute_uv:
            U_loc, S_loc, Vt_loc = torch.linalg.svd(A.larray, full_matrices=full_matrices)
            U = _toDNDarray(U_loc)
            S = _toDNDarray(S_loc)
            Vt = _toDNDarray(Vt_loc)
            return U, S, Vt
        else:
            S_loc = torch.linalg.svdvals(A.larray)
            S = _toDNDarray(S_loc)
            return S
    elif A.split == 0 and A.lshape_map[:, 0].max().item() >= A.shape[1]:
        # this is the distributed, tall skinny case
        # compute SVD via tall skinny QR
        if compute_uv:
            # compute full SVD: first full QR, then SVD of R
            Q, R = qr(A, mode="reduced", procs_to_merge=qr_procs_to_merge)
            Utilde_loc, S_loc, Vt_loc = torch.linalg.svd(R.larray, full_matrices=False)
            Utilde = _toDNDarray(Utilde_loc)
            S = _toDNDarray(S_loc)
            Vt = _toDNDarray(Vt_loc)
            U = (Utilde.T @ Q.T).T
            return U, S, Vt
        else:
            # compute only singular values: first only R of QR, then singular values only of R
            _, R = qr(A, mode="r", procs_to_merge=qr_procs_to_merge)
            S_loc = torch.linalg.svdvals(R.larray)
            S = _toDNDarray(S_loc)
            return S
    elif A.split == 1 and A.lshape_map[:, 1].max().item() >= A.shape[0]:
        # this is the distributed, short fat case
        # apply the tall skinny SVD to the transpose of A
        if compute_uv:
            V, S, Ut = svd(
                A.T,
                full_matrices=full_matrices,
                compute_uv=True,
                qr_procs_to_merge=qr_procs_to_merge,
            )
            return Ut.T, S, V.T
        else:
            S = svd(
                A.T,
                full_matrices=full_matrices,
                compute_uv=False,
                qr_procs_to_merge=qr_procs_to_merge,
            )
            return S

    else:
        # this is the general, distributed case in which the matrix is neither tall skinny nor short fat
        # we apply the Zolotarev-Polar Decomposition and the symmetric eigenvalue decomposition
        if A.shape[0] < A.shape[1]:
            # Zolo-PD requires A.shape[0] >= A.shape[1], so we need to transpose in this case
            if compute_uv:
                V, S, Ut = svd(
                    A.T,
                    full_matrices=full_matrices,
                    compute_uv=True,
                    qr_procs_to_merge=qr_procs_to_merge,
                )
                return Ut.T, S, V.T
            else:
                S = svd(
                    A.T,
                    full_matrices=full_matrices,
                    compute_uv=False,
                    qr_procs_to_merge=qr_procs_to_merge,
                )
                return S
        else:
            warn(
                "You are performing the full SVD of a distributed matrix that is neither of tall-skinny nor short-fat shape. \n This operation may be costly in terms of memory and compute time."
            )
            U, H = polar(A, r_max=r_max_zolopd)
            S, V = eigh(H, r_max_zolopd=r_max_zolopd)
            if not compute_uv:
                return S
            else:
                return U @ V, S, V.T



import heat as ht
import torch
from heat.core.dndarray import DNDarray
from warnings import warn


from typing import Union, Tuple, TypeVar

self = TypeVar("self")

import numpy as np
# from ..core import BaseEstimator  # Assuming BaseEstimator and ClusteringMixin are in a parent or sibling directory in a library structure.
# from ..utils import ClusteringMixin
from typing import Union, List, Tuple, Dict # Import necessary type hints

import itertools # For generating combinations (potentially)

class Dendrogram(): # Removed inheritance from BaseEstimator and ClusteringMixin for simplicity in standalone context, can be added back if needed

    def __init__(
        self,
        connectivity: int = 4, # Default connectivity as 4 (2D) or 6 (3D)
        dmax: int = 3, # Spatial box size parameter
        vmax: int = 7, # Velocity/Channel box size parameter (for 3D)
        rms_noise: float = 0.3, # RMS noise level
        ppb: float = 1.3, # Pixels per beam
        min_value: Union[float, None] = None, # Minimum value threshold
        enforce_binary_mergers: bool = True, # Enforce binary mergers (basic check, refinement not fully implemented)
    ):
        """
        Initializes the Dendrogram object with parameters for dendrogram construction.

        Args:
            connectivity (int): Connectivity type (4, 8 in 2D; 6, 18, 26 in 3D).
            dmax (int): Spatial box size parameter for local maxima refinement.
            vmax (int): Velocity/Channel box size parameter for local maxima refinement (3D).
            rms_noise (float): Estimated RMS noise level.
            ppb (float): Pixels per beam, for minimum pixel volume criterion.
            min_value (float, None): Intensity value below which to ignore (noise threshold).
            enforce_binary_mergers (bool): Whether to perform basic check for binary mergers.
        """
        self.connectivity = connectivity
        self.dmax = dmax
        self.vmax = vmax
        self.rms_noise = rms_noise
        self.ppb = ppb
        self.min_value = min_value
        self.enforce_binary_mergers = enforce_binary_mergers
        self.name = "Vectorized Dendrogram" # Name to distinguish implementation


    def compute(self, image: torch.Tensor, contour_levels: Union[List[float], np.ndarray]) -> Tuple[Dict, List, torch.Tensor]: # Specify return type
        """
        Computes the dendrogram for a given intensity image and contour levels.

        Args:
            image (torch.Tensor): N-dimensional intensity image.
            contour_levels (list or np.ndarray): List of contour levels (descending order).

        Returns:
            tuple: (dendrogram, merge_history, local_maxima_mask).
                dendrogram (dict): Dendrogram information (merges at each level).
                merge_history (list): List of all merge events.
                local_maxima_mask (torch.Tensor): Boolean mask of decimated local maxima.
        """
        dendrogram_info, merge_history_info, local_maxima_mask_info = self._build_dendrogram_refined_maxima_internal(image, contour_levels) # Call internal method
        return dendrogram_info, merge_history_info, local_maxima_mask_info


    def _build_dendrogram_refined_maxima_internal(self, image, contour_levels): # Renamed and made internal
        """
        Internal method to construct a dendrogram with refined local maxima.
        """
        dendrogram = {}
        previous_regions_mask = None
        merge_history = []

        tmax_factor = 2.0 # Based on astrodendro example: min_delta=2*sigma
        nmin = int(3 * self.ppb) # Based on astrodendro example: min_npix=3*ppb

        # **Image Pre-processing: Apply min_value threshold**
        if self.min_value is not None:
            image_masked = image.clone() #clone to avoid in-place modification
            image_masked[image_masked < self.min_value] = torch.nan # Set values below min_value to NaN
        else:
            image_masked = image # If min_value is None, use original image


        # Use refined local maxima identification - Handles NaNs, now on masked image
        local_maxima_mask = self._identify_local_maxima_refined( # Call internal method
            image_masked, contour_levels, self.dmax, self.vmax, tmax_factor, nmin, self.connectivity, self.rms_noise
        )

        # Original dendrogram building logic (modified to potentially use local_maxima_mask if needed) - operates on masked image
        for level in contour_levels:
            binary_mask = self._contour_image_internal(image_masked, level) # contour_image handles NaNs, now on masked image # Call internal method
            current_regions_mask, num_regions = self._identify_regions_internal(binary_mask, self.connectivity) # identify_regions (BFS) handles NaNs, on masked image # Call internal method

            if previous_regions_mask is not None:
                merges = self._detect_merges_internal(previous_regions_mask, current_regions_mask) # detect_merges handles NaNs in region masks # Call internal method
                if merges:
                    print(f"Merges detected at contour level: {level}, Merged regions (previous level IDs): {merges}")
                    dendrogram[level] = merges
                    merge_history.extend([(level, merge) for merge in merges])

                    if self.enforce_binary_mergers: # Basic check - not full refinement
                        for merge_group in merges:
                            if len(merge_group) > 2:
                                print(f"Warning: Non-binary merge detected at level {level} involving regions {merge_group}. Binary merger enforcement is not fully implemented; level refinement is needed.")


            previous_regions_mask = current_regions_mask

        return dendrogram, merge_history, local_maxima_mask # Return also the local maxima mask if needed later


    def _identify_local_maxima_refined(self, image, contour_levels, dmax, vmax, tmax_factor, nmin, connectivity, rms_noise): # Made internal
        """
        Internal method - Identifies local maxima with noise suppression using box filter and decimation.
        """
        # 1. Identify Candidate Local Maxima (Box-based) - Handles NaNs
        candidate_maxima_mask = self._identify_candidate_local_maxima_box_internal(image, dmax, vmax) # Call internal method
        candidate_maxima_coords_list = []
        for coords_index in np.ndindex(image.shape):
            if candidate_maxima_mask[coords_index]:
                candidate_maxima_coords_list.append(coords_index)

        print(f"Number of candidate local maxima (before decimation): {len(candidate_maxima_coords_list)}")

        # 2. Decimate Local Maxima based on noise criteria - Handles NaNs in merge level and volume calculation
        device = image.device # Get device of the image tensor and pass it to decimate_local_maxima
        decimated_maxima_coords_list = self._decimate_local_maxima_internal( # Call internal method
            image, candidate_maxima_coords_list, contour_levels, dmax, vmax, tmax_factor, nmin, connectivity, rms_noise
        )

        print(f"Number of decimated local maxima (after noise suppression): {len(decimated_maxima_coords_list)}")

        # Create a boolean mask for the decimated maxima
        decimated_maxima_mask = torch.zeros_like(image, dtype=torch.bool, device=image.device) # Ensure mask is on same device as image
        for coords in decimated_maxima_coords_list:
            decimated_maxima_mask[coords] = True

        return decimated_maxima_mask


    def _decimate_local_maxima_internal(self, image, candidate_maxima_coords_list, contour_levels, dmax, vmax, tmax_factor, nmin, connectivity, rms_noise): # Made internal
        """
        Vectorized Internal method - Decimates candidate local maxima based on noise suppression criteria.
        --- VECTORIZED VERSION ---
        """
        tmax = tmax_factor * rms_noise # Tmax threshold based on noise level
        decimated_maxima_coords = list(candidate_maxima_coords_list) # Start with all candidates - list of tuples

        if not decimated_maxima_coords or len(decimated_maxima_coords) < 2: # Need at least 2 maxima to compare
            return decimated_maxima_coords


        maxima_coords_tensor = torch.tensor(decimated_maxima_coords, dtype=torch.long, device=image.device) # Convert maxima coords to tensor [N_maxima, ndim]
        num_maxima = maxima_coords_tensor.shape[0]

        # 1. Generate all pairs of maxima indices
        maxima_indices = torch.arange(num_maxima) # [0, 1, 2, ..., N_maxima-1]
        pairs_indices = torch.combinations(maxima_indices, r=2) # Get all unique pairs of indices [N_pairs, 2]

        num_pairs = pairs_indices.shape[0]
        if num_pairs == 0: # No pairs to compare
            return decimated_maxima_coords


        maxima1_indices = pairs_indices[:, 0] # Indices of the first maximum in each pair [N_pairs]
        maxima2_indices = pairs_indices[:, 1] # Indices of the second maximum in each pair [N_pairs]


        current_maxima_coords_tensor = maxima_coords_tensor # Working copy - will be filtered

        maxima1_coords = current_maxima_coords_tensor[maxima1_indices] # Coordinates of the first maxima in each pair [N_pairs, ndim]
        maxima2_coords = current_maxima_coords_tensor[maxima2_indices] # Coordinates of the second maxima in each pair [N_pairs, ndim]


        # Initialize masks to keep track of maxima to remove (initially none to remove)
        remove_maxima_mask = torch.zeros(num_maxima, dtype=torch.bool, device=image.device) # [N_maxima] - boolean mask to mark maxima for removal



        for level in contour_levels: # Iterate through contour levels (descending)

            # --- Vectorized Merge Check and Volume Calculation ---
            merge_info_list = self._calculate_merge_level_and_volume_vectorized_internal(image, maxima1_coords, maxima2_coords, level, connectivity, nmin) # Vectorized call

            merge_levels_tensor, vol1_tensor, vol2_tensor, merged_mask_level = merge_info_list # [N_pairs], [N_pairs], [N_pairs], [N_pairs] (bool) - for the *current level*


            # --- Apply Noise Criteria (Vectorized) - only to pairs that *did* merge at this level ---
            merged_pairs_indices_level = pairs_indices[merged_mask_level] # Indices of pairs that merged at *this level* [N_merged_pairs_level, 2]
            if merged_pairs_indices_level.numel() == 0: # No merges at this level
                continue # Go to next contour level


            merged_maxima1_indices_level = merged_pairs_indices_level[:, 0] # Indices of max1 for merged pairs at this level [N_merged_pairs_level]
            merged_maxima2_indices_level = merged_pairs_indices_level[:, 1] # Indices of max2 for merged pairs at this level [N_merged_pairs_level]


            merged_vol1_level = vol1_tensor[merged_mask_level] # Volumes for merged pairs [N_merged_pairs_level]
            merged_vol2_level = vol2_tensor[merged_mask_level] # Volumes for merged pairs [N_merged_pairs_level]
            merged_level_proxy_level = merge_levels_tensor[merged_mask_level] # Merge levels (proxy) [N_merged_pairs_level]


            deltaT1_level = image[tuple(maxima1_coords[merged_maxima1_indices_level].T)] - merged_level_proxy_level # [N_merged_pairs_level]
            deltaT2_level = image[tuple(maxima2_coords[merged_maxima2_indices_level].T)] - merged_level_proxy_level # [N_merged_pairs_level]
            tmax_tensor = torch.full_like(deltaT1_level, tmax) # [N_merged_pairs_level] - tensor of tmax values


            # --- Volume Criterion (Vectorized) ---
            volume_criterion_failed_mask = (merged_vol1_level < nmin) | (merged_vol2_level < nmin) # [N_merged_pairs_level]

            # --- Contrast Criterion (Vectorized) ---
            contrast_criterion_failed_mask = (deltaT1_level < tmax_tensor) | (deltaT2_level < tmax_tensor) # [N_merged_pairs_level]


            # --- Determine Maxima to Remove (Vectorized) ---
            maxima_to_remove_indices_level_max1 = merged_maxima1_indices_level[volume_criterion_failed_mask & (merged_vol2_level >= nmin)] # Max1 to remove if vol1 < nmin and vol2 >= nmin
            maxima_to_remove_indices_level_max2 = merged_maxima2_indices_level[volume_criterion_failed_mask & (merged_vol1_level >= nmin)] # Max2 to remove if vol2 < nmin and vol1 >= nmin

            # Both volumes small, remove lower intensity one (vectorized)
            both_volumes_small_mask = volume_criterion_failed_mask & (merged_vol1_level < nmin) & (merged_vol2_level < nmin) # Both vols < nmin
            lower_intensity_maxima_indices_max1 = merged_maxima1_indices_level[both_volumes_small_mask & (image[tuple(maxima1_coords[merged_maxima1_indices_level].T)] <= image[tuple(maxima2_coords[merged_maxima2_indices_level].T)])] # if max1 intensity <= max2
            lower_intensity_maxima_indices_max2 = merged_maxima2_indices_level[both_volumes_small_mask & (image[tuple(maxima1_coords[merged_maxima1_indices_level].T)] > image[tuple(maxima2_coords[merged_maxima2_indices_level].T)])] # if max1 intensity > max2


            # Contrast Criterion failures - similar logic as volume
            contrast_failed_maxima_indices_max1 = merged_maxima1_indices_level[contrast_criterion_failed_mask & (deltaT2_level >= tmax_tensor)] # Max1 to remove if deltaT1 < tmax and deltaT2 >= tmax
            contrast_failed_maxima_indices_max2 = merged_maxima2_indices_level[contrast_criterion_failed_mask & (deltaT1_level >= tmax_tensor)] # Max2 to remove if deltaT2 < tmax and deltaT1 >= tmax

            # Both contrasts fail - remove lower intensity one
            both_contrasts_fail_mask = contrast_criterion_failed_mask & (deltaT1_level < tmax_tensor) & (deltaT2_level < tmax_tensor)
            lower_intensity_contrast_fail_maxima_indices_max1 = merged_maxima1_indices_level[both_contrasts_fail_mask & (image[tuple(maxima1_coords[merged_maxima1_indices_level].T)] <= image[tuple(maxima2_coords[merged_maxima2_indices_level].T)])]
            lower_intensity_contrast_fail_maxima_indices_max2 = merged_maxima2_indices_level[both_contrasts_fail_mask & (image[tuple(maxima1_coords[merged_maxima1_indices_level].T)] > image[tuple(maxima2_coords[merged_maxima2_indices_level].T)])]


            # --- Update remove_maxima_mask (Vectorized) ---
            remove_maxima_mask[maxima_to_remove_indices_level_max1] = True
            remove_maxima_mask[maxima_to_remove_indices_level_max2] = True
            remove_maxima_mask[lower_intensity_maxima_indices_max1] = True
            remove_maxima_mask[lower_intensity_maxima_indices_max2] = True
            remove_maxima_mask[contrast_failed_maxima_indices_max1] = True
            remove_maxima_mask[contrast_failed_maxima_indices_max2] = True
            remove_maxima_mask[lower_intensity_contrast_fail_maxima_indices_max1] = True
            remove_maxima_mask[lower_intensity_contrast_fail_maxima_indices_max2] = True


        # --- Filter Maxima Coordinates (Vectorized) ---
        keep_maxima_mask = ~remove_maxima_mask # Invert the mask to keep maxima that are *not* marked for removal
        decimated_maxima_coords_tensor_final = current_maxima_coords_tensor[keep_maxima_mask] # Apply mask to filter

        decimated_maxima_coords_list = [tuple(coords) for coords in decimated_maxima_coords_tensor_final.tolist()] # Convert back to list of tuples


        return decimated_maxima_coords_list



    def _calculate_merge_level_and_volume_vectorized_internal(self, image, maxima1_coords, maxima2_coords, contour_level, connectivity, nmin):
        """
        Vectorized Internal method - Calculates merge level and volumes for *pairs* of local maxima.
        """
        merge_level_proxy = contour_level # Using contour level as proxy for merge level
        num_pairs = maxima1_coords.shape[0]

        # 1. Contour at the given level (once for all pairs at this level)
        binary_mask = self._contour_image_internal(image, merge_level_proxy) # Call internal method
        regions_mask, num_regions = self._identify_regions_internal(binary_mask, connectivity) # Call internal method

        # 2. Get Region Labels for Maxima (Vectorized)
        maxima1_region_labels = regions_mask[tuple(maxima1_coords.T)] # [N_pairs] - region labels for max1 in each pair
        maxima2_region_labels = regions_mask[tuple(maxima2_coords.T)] # [N_pairs] - region labels for max2 in each pair


        # 3. Check for Merges (Vectorized) - Maxima in same non-background region
        merged_mask = ~(torch.isnan(maxima1_region_labels) | torch.isnan(maxima2_region_labels)) & (maxima1_region_labels != 0) & (maxima1_region_labels == maxima2_region_labels) # [N_pairs] - boolean mask - True if merged


        merge_levels_tensor = torch.full((num_pairs,), merge_level_proxy, dtype=image.dtype, device=image.device) # [N_pairs] - tensor of merge levels (proxy)
        vol1_tensor = torch.zeros(num_pairs, dtype=torch.float, device=image.device) # [N_pairs] - volumes for max1's regions
        vol2_tensor = torch.zeros(num_pairs, dtype=torch.float, device=image.device) # [N_pairs] - volumes for max2's regions


        # 4. Calculate Volumes for Merged Pairs (Vectorized - only for merged pairs)
        merged_pair_indices = torch.arange(num_pairs)[merged_mask] # Indices of pairs that are merged [N_merged_pairs_level]

        if merged_pair_indices.numel() > 0: # Only compute volumes if there are merged pairs
            merged_region_labels = maxima1_region_labels[merged_mask] # Region labels for merged pairs (using max1 labels as they are same) [N_merged_pairs_level]

            for i in range(merged_pair_indices.shape[0]): # Iterate through merged pair indices (can we vectorize this further?)
                pair_index = merged_pair_indices[i] # Original pair index
                region_label = merged_region_labels[i].int() # Get int label for masking
                merged_region_mask = (regions_mask == region_label) # Mask of the merged region

                volume = torch.sum(merged_region_mask).item() # Calculate volume for the *merged* region (shared volume)
                vol1_tensor[pair_index] = volume # Assign same volume to both for now - simplification
                vol2_tensor[pair_index] = volume # Assign same volume to both for now - simplification


        return merge_levels_tensor, vol1_tensor, vol2_tensor, merged_mask # Return all info as tensors



    def _identify_candidate_local_maxima_box_internal(self, image, dmax, vmax): # Made internal
        """
        Internal method - Identifies candidate local maxima using a box filter approach.
        Compatibility fix for PyTorch versions < 1.12 (no nan_to_num in MaxPool).
        """
        ndim = image.ndim
        if ndim == 2:
            kernel_size = 2 * dmax + 1
            padding = dmax
            max_pool = torch.nn.MaxPool2d(kernel_size=kernel_size, padding=padding) # nan_to_num removed for older PyTorch
        elif ndim == 3:
            kernel_size = (2 * vmax + 1, 2 * dmax + 1, 2 * dmax + 1) # (depth, height, width)
            padding = (vmax, dmax, dmax)
            print(f"kernel_size: {kernel_size}, padding: {padding}")
            max_pool = torch.nn.MaxPool3d(kernel_size=kernel_size, padding=padding) # nan_to_num removed for older PyTorch
        else:
            raise ValueError("Image must be 2D or 3D.")

        # Replace NaNs with a very small negative number *before* max pooling for older PyTorch versions
        nan_mask = torch.isnan(image)
        print(f"Number of NaNs in image: {torch.sum(nan_mask)}")
        image_nan_filled = image.clone() # Clone to avoid modifying original image
        image_nan_filled[nan_mask] = torch.tensor(float('-inf'), dtype=image.dtype, device=image.device) # Replace NaNs with -inf
        max_pooled_image = max_pool(image_nan_filled.unsqueeze(0).unsqueeze(0)).squeeze(0).squeeze(0) # Add channel and batch dims for pooling

        # Compare original image with max-pooled image to find maxima candidates
        candidate_maxima = (image == max_pooled_image)
        return candidate_maxima


    def _detect_merges_internal(self, previous_regions_mask, current_regions_mask): # Made internal
        """
        Internal method - Detects merges between regions from a previous contour level to the current one.
        """
        if previous_regions_mask is None: # No previous regions to compare to at the very first level
            return []

        merges = []
        previous_region_labels = set(int(x) for x in torch.unique(previous_regions_mask) if x != 0 and not torch.isnan(x)) # Exclude background label 0 and NaN labels
        current_region_labels = set(int(x) for x in torch.unique(current_regions_mask) if x != 0 and not torch.isnan(x)) # Exclude background label 0 and NaN labels


        # For each region in the previous level, find which current region(s) it overlaps with
        region_overlap = {} # Key: previous region label, Value: set of current region labels it overlaps with

        for prev_label in previous_region_labels:
            overlap_current_labels = set()
            prev_region_mask = (previous_regions_mask == prev_label)
            overlap_mask = prev_region_mask & (current_regions_mask != 0) # Pixels of prev region that are also in any current region (non-background)
            overlapping_current_region_labels = set(int(x) for x in torch.unique(current_regions_mask[overlap_mask]) if x != 0 and not torch.isnan(x)) # Exclude background label 0 and NaN labels
            region_overlap[prev_label] = overlapping_current_region_labels

        merged_region_map = {} # Key: current region label, Value: set of previous region labels merged into it

        for prev_label in previous_region_labels:
            for current_label in region_overlap[prev_label]:
                merged_region_map.setdefault(current_label, set()).add(prev_label)

        detected_merges = []
        for current_label, prev_labels in merged_region_map.items():
            if len(prev_labels) > 1: # More than one previous region merged into this current region - a merge occurred.
                detected_merges.append(tuple(sorted(list(prev_labels)))) # Store the merged previous region IDs (tuple for immutability)

        region_merges_list = []
        for current_label in current_region_labels:
            current_region_mask = (current_regions_mask == current_label)
            overlapping_prev_region_labels = set(int(x) for x in torch.unique(previous_regions_mask[current_region_mask]) if x != 0 and not torch.isnan(x)) # Exclude background label 0 and NaN labels
            if len(overlapping_prev_region_labels) > 1:
                region_merges_list.append(tuple(sorted(list(overlapping_prev_region_labels)))) # Store merged previous region IDs


        # Remove duplicates and sort the merges (by merged region IDs)
        unique_merges = sorted(list(set(region_merges_list))) # Use set to remove duplicates, then sort for consistent order

        return unique_merges


    def _identify_regions_internal(self, binary_contour_mask, connectivity): # Made internal
        """
        Internal method - Identifies connected regions in a binary contour mask using connected components labeling.
        """
        return self._identify_regions_torch_bfs_internal(binary_contour_mask, connectivity) # Call internal method


    def _identify_regions_torch_bfs_internal(self, binary_contour_mask, connectivity): # Made internal
        """
        Internal method - Identifies connected regions in a binary contour mask using BFS (PyTorch implementation for GPU).
        """
        image_shape = binary_contour_mask.shape
        labeled_mask = torch.zeros_like(binary_contour_mask, dtype=torch.int32, device=binary_contour_mask.device) # Initialize labeled mask on same device
        region_label = 0

        # Get indices of foreground pixels (True values, excluding NaNs - NaNs in boolean context are False)
        foreground_indices = (binary_contour_mask > 0).nonzero() # Get coordinates of foreground pixels

        processed_pixels = torch.zeros_like(binary_contour_mask, dtype=torch.bool, device=binary_contour_mask.device) # Track processed pixels on same device

        for start_coords_index in foreground_indices:
            start_coords = tuple(start_coords_index.tolist()) # Convert to tuple for indexing

            if not processed_pixels[start_coords]: # If pixel not yet processed, start new region
                region_label += 1
                queue = [start_coords] # Initialize BFS queue
                processed_pixels[start_coords] = True
                labeled_mask[start_coords] = region_label

            while queue:
                current_coords = queue.pop(0)

                for neighbor_coords_tensor in self._get_neighbors_torch_internal(current_coords, image_shape, connectivity): # Get neighbors as tensor # Call internal method
                    neighbor_coords = tuple(neighbor_coords_tensor.tolist()) # Convert neighbor coords to tuple for indexing

                    if (0 <= neighbor_coords[0] < image_shape[0] and 0 <= neighbor_coords[1] < image_shape[1] and  # Basic bounds check - already done in get_neighbors_torch, but double check for 2D
                        binary_contour_mask[neighbor_coords] and not processed_pixels[neighbor_coords]): # Check if neighbor is foreground and not processed (binary_contour_mask[NaN] is False)

                        processed_pixels[neighbor_coords] = True
                        labeled_mask[neighbor_coords] = region_label
                        queue.append(neighbor_coords) # Add neighbor to queue

        num_regions = region_label # Max region label is the number of regions

        return labeled_mask, num_regions


    def _contour_image_internal(self, image, level): # Made internal
        """
        Internal method - Contours the image at a given intensity level.
        """
        return image >= level


    def _get_neighbors_torch_internal(self, coords, image_shape, connectivity): # Made internal, static
        """
        Internal static method - Get neighbors of a pixel/voxel based on connectivity, PyTorch version.
        """
        ndim = len(image_shape)
        neighbor_offsets_list = []

        if ndim == 2:
            if connectivity == 4:
                neighbor_offsets_list = [(-1, 0), (1, 0), (0, -1), (0, 1)]
            elif connectivity == 8:
                neighbor_offsets_list = [(-1, -1), (-1, 0), (-1, 1), (0, -1), (0, 1), (1, -1), (1, 0), (1, 1)]
            else:
                raise ValueError("Invalid connectivity for 2D. Choose 4 or 8.")
        elif ndim == 3:
            if connectivity == 6:
                neighbor_offsets_list = [(-1, 0, 0), (1, 0, 0), (0, -1, 0), (0, 1, 0), (0, 0, -1), (0, 0, 1)]
            elif connectivity == 26:
                neighbor_offsets_list = []
                for dd in [-1, 0, 1]:
                    for dr in [-1, 0, 1]:
                        for dc in [-1, 0, 1]:
                            if (dd, dr, dc) != (0, 0, 0):
                                neighbor_offsets_list.append((dd, dr, dc))
            elif connectivity == 18:
                neighbor_offsets_list = [
                    (-1, 0, 0), (1, 0, 0), (0, -1, 0), (0, 1, 0), (0, 0, -1), (0, 0, 1),
                    (-1, -1, 0), (-1, 1, 0), (-1, 0, -1), (-1, 0, 1),
                    (1, -1, 0), (1, 1, 0), (1, 0, -1), (1, 0, 1),
                    (0, -1, -1), (0, -1, 1), (0, 1, -1), (0, 1, 1)
                ]
            elif connectivity == 6:
                 neighbor_offsets_list = [(-1, 0, 0), (1, 0, 0), (0, -1, 0), (0, 1, 0), (0, 0, -1), (0, 0, 1)]
            else:
                raise ValueError("Invalid connectivity for 3D. Choose 6, 18, or 26.")
        else:
            raise ValueError("Image must be 2D or 3D.")

        neighbor_offsets = torch.tensor(neighbor_offsets_list, dtype=torch.long) # Convert to tensor

        coords_tensor = torch.tensor(coords, dtype=torch.long) # Input coords to tensor

        neighbor_coords = coords_tensor + neighbor_offsets # Calculate neighbor coords

        # Keep only valid coordinates within image bounds
        valid_neighbors_mask = torch.ones(neighbor_coords.shape[0], dtype=torch.bool)
        for i in range(ndim):
            valid_neighbors_mask = valid_neighbors_mask & (neighbor_coords[:, i] >= 0) & (neighbor_coords[:, i] < image_shape[i])

        return neighbor_coords[valid_neighbors_mask] # Return only valid neighbors


# Re-assigning methods to the class (no changes in these, just ensuring they are part of the Dendrogram class)
Dendrogram._identify_candidate_local_maxima_box_internal = Dendrogram._identify_candidate_local_maxima_box_internal
Dendrogram._detect_merges_internal = Dendrogram._detect_merges_internal
Dendrogram._identify_regions_internal = Dendrogram._identify_regions_internal
Dendrogram._identify_regions_torch_bfs_internal = Dendrogram._identify_regions_torch_bfs_internal
Dendrogram._contour_image_internal = Dendrogram._contour_image_internal
Dendrogram._get_neighbors_torch_internal = Dendrogram._get_neighbors_torch_internal


if __name__ == '__main__':
    # Example Usage and Comparison - Including original loop-based Dendrogram for timing comparison
    import time

    # Dummy 3D data with a peak and some NaNs (adjust size for timing tests)
    data_3d_np = np.zeros((30, 60, 60), dtype=np.float32) # Larger data for timing
    data_3d_np[15, 30, 30] = 20  # Strong peak
    data_3d_np[10:20, 20:40, 20:40] = 10 # Extended emission
    data_3d_np[:5, :, :] = np.nan  # Add NaN slices at the beginning
    data_3d_np[:, :10, :] = np.nan  # Add NaN columns at the beginning
    data_cube_torch = torch.from_numpy(data_3d_np).to("cpu") # Run on CPU for consistent timing in example


    sigma_noise = 0.5 # Example noise level
    pixels_per_beam = 1.5 # Example pixels per beam
    min_value_threshold_K = 0.75 # Example min_value
    contour_levels_example = np.linspace(np.nanmax(data_3d_np), min_value_threshold_K, num=10)


    # 1. Original Loop-Based Dendrogram (for timing comparison - assuming you have the loop-based version saved as 'dendrogram_loop.py')
    # (You would need to have a Dendrogram class saved in 'dendrogram_loop.py' with the original _decimate_local_maxima_internal)
    # from dendrogram_loop import Dendrogram as DendrogramLoop # Uncomment if you have the loop-based version in a separate file
    class DendrogramLoop(Dendrogram): # Temporary - rename to DendrogramLoop and adjust name in init if you want to keep original code in same file
        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)
            self.name = "Loop-Based Dendrogram" # For comparison output
        _decimate_local_maxima_internal = Dendrogram.__dict__['_decimate_local_maxima_internal'] # Revert to loop-based version (if you have it saved under the same class name)
        # (If you have a separate file, just import DendrogramLoop from it instead of this temporary definition)


    dendrogram_loop_class = DendrogramLoop(connectivity=6, dmax=2, vmax=2, rms_noise=sigma_noise, ppb=pixels_per_beam, min_value=min_value_threshold_K)


    start_time_loop = time.perf_counter()
    dendrogram_loop_output, merge_history_loop_output, local_maxima_mask_loop_output = dendrogram_loop_class.compute(data_cube_torch, contour_levels_example)
    end_time_loop = time.perf_counter()
    time_loop = end_time_loop - start_time_loop

    num_maxima_loop = int(torch.sum(local_maxima_mask_loop_output).item())


    # 2. Vectorized Dendrogram Class (the code from this response)
    dendrogram_vectorized_class = Dendrogram(connectivity=6, dmax=2, vmax=2, rms_noise=sigma_noise, ppb=pixels_per_beam, min_value=min_value_threshold_K)

    start_time_vectorized = time.perf_counter()
    dendrogram_vectorized_output, merge_history_vectorized_output, local_maxima_mask_vectorized_output = dendrogram_vectorized_class.compute(data_cube_torch, contour_levels_example)
    end_time_vectorized = time.perf_counter()
    time_vectorized = end_time_vectorized - start_time_vectorized

    num_maxima_vectorized = int(torch.sum(local_maxima_mask_vectorized_output).item())


    # --- Comparison Output ---
    print("\n--- Dendrogram Implementation Comparison (Timing and Maxima Count) ---")
    print(f"Data Cube Size: {data_3d_np.shape}")
    print(f"Contour Levels: {len(contour_levels_example)}")
    print(f"Parameters: connectivity={6}, dmax=2, vmax=2, rms_noise={sigma_noise}, ppb={pixels_per_beam}, min_value={min_value_threshold_K}")

    print(f"\n{dendrogram_loop_class.name}:")
    print(f"  Execution Time: {time_loop:.4f} seconds")
    print(f"  Number of Maxima: {num_maxima_loop}")

    print(f"\n{dendrogram_vectorized_class.name}:")
    print(f"  Execution Time: {time_vectorized:.4f} seconds")
    print(f"  Number of Maxima: {num_maxima_vectorized}")

    speedup = time_loop / time_vectorized if time_vectorized > 0 else float('inf')
    maxima_diff_percent = 0
    if max(num_maxima_loop, num_maxima_vectorized) > 0:
        maxima_diff_percent = abs(num_maxima_loop - num_maxima_vectorized) / max(num_maxima_loop, num_maxima_vectorized) * 100.0

    print(f"\nSpeedup (Loop-based vs. Vectorized): ~{speedup:.2f}x")
    print(f"Difference in Number of Maxima: {abs(num_maxima_loop - num_maxima_vectorized)} ( ~{maxima_diff_percent:.2f}%)")

    # --- (Optional) Add more detailed comparison of dendrogram structure and merge history if needed ---
    print("\n--- Comparison Complete ---")



"""
Everything you need to know about the configuration of Heat
"""

import torch
import platform
import mpi4py
import subprocess
import os
import warnings
import re
import dataclasses
from enum import Enum


class MPILibrary(Enum):
    OpenMPI = "ompi"
    IntelMPI = "impi"
    MVAPICH = "mvapich"
    MPICH = "mpich"
    CrayMPI = "craympi"
    ParaStationMPI = "psmpi"
    Other = "other"


@dataclasses.dataclass
class MPILibraryInfo:
    name: MPILibrary
    version: str


def _get_mpi_library() -> MPILibraryInfo:
    library = mpi4py.MPI.Get_library_version().split()
    match library:
        case ["Open", "MPI", *_]:
            return MPILibraryInfo(MPILibrary.OpenMPI, library[2])
        case ["Intel(R)", "MPI", *_]:
            return MPILibraryInfo(MPILibrary.IntelMPI, library[3])
        case ["MPICH", "Version:", *_]:
            return MPILibraryInfo(MPILibrary.MPICH, library[2])
        case ["MVAPICH", "Version:", *_]:
            return MPILibraryInfo(MPILibrary.MVAPICH, library[2])
        case ["===", "ParaStation", "MPI", *_]:
            return MPILibraryInfo(MPILibrary.ParaStationMPI, library[3])
        case _:
            return MPILibraryInfo(MPILibrary.Other, "unknown")


def _check_gpu_aware_mpi(library: MPILibraryInfo) -> tuple[bool, bool]:
    match library.name:
        case MPILibrary.OpenMPI:
            try:
                parsable_ompi_info = subprocess.check_output(
                    ["ompi_info", "--parsable", "--all"]
                ).decode("utf-8")
                ompi_info = subprocess.check_output(["ompi_info"]).decode("utf-8")

                # Check for CUDA support flag
                cuda_support_flag = "mpi_built_with_cuda_support:value:true" in parsable_ompi_info

                # Check for extensions
                match = re.search(r"MPI extensions: (.*)", ompi_info)
                extensions = [ext.strip() for ext in match.group(0).split(":")[1].split(",")]
                cuda = cuda_support_flag and "cuda" in extensions
                if library.version.startswith("v4."):
                    rocm = cuda
                elif library.version.startswith("v5."):
                    rocm = "rocm" in extensions or "hip" in extensions
                # Seems to be broken, disabled by default for now
                # return cuda, rocm
                return False, False
            except Exception as e:  # noqa E722
                return False, False
        case MPILibrary.IntelMPI:
            return False, False
        case MPILibrary.MVAPICH:
            cuda = os.environ.get("MV2_USE_CUDA") == "1"
            rocm = os.environ.get("MV2_USE_ROCM") == "1"
            return cuda, rocm
        case MPILibrary.MPICH:
            cuda = os.environ.get("MPIR_CVAR_ENABLE_HCOLL") == "1"
            rocm = False
            return cuda, rocm
        case MPILibrary.CrayMPI:
            cuda = os.environ.get("MPICH_GPU_SUPPORT_ENABLED") == "1"
            rocm = os.environ.get("MPICH_GPU_SUPPORT_ENABLED") == "1"
            return cuda, rocm
        case MPILibrary.ParaStationMPI:
            cuda = os.environ.get("PSP_CUDA") == "1"
            rocm = False
            return cuda, rocm
        case _:
            return False, False


PLATFORM = platform.platform()
TORCH_VERSION = torch.__version__
TORCH_CUDA_IS_AVAILABLE = torch.cuda.is_available()
CUDA_IS_ACTUALLY_ROCM = "rocm" in TORCH_VERSION

mpi_library = _get_mpi_library()
CUDA_AWARE_MPI, ROCM_AWARE_MPI = _check_gpu_aware_mpi(mpi_library)
GPU_AWARE_MPI = False

# warn the user if CUDA/ROCm-aware MPI is not available, but PyTorch can use GPUs with CUDA/ROCm
if TORCH_CUDA_IS_AVAILABLE:
    if not CUDA_IS_ACTUALLY_ROCM and not CUDA_AWARE_MPI:
        warnings.warn(
            f"Heat has CUDA GPU-support (PyTorch version {TORCH_VERSION} and `torch.cuda.is_available() = True`), but CUDA-awareness of MPI could not be detected. This may lead to performance degradation as direct MPI-communication between GPUs is not possible.",
            UserWarning,
        )

    elif CUDA_IS_ACTUALLY_ROCM and not ROCM_AWARE_MPI:
        warnings.warn(
            f"Heat has ROCm GPU-support (PyTorch version {TORCH_VERSION} and `torch.cuda.is_available() = True`), but ROCm-awareness of MPI could not be detected. This may lead to performance degradation as direct MPI-communication between GPUs is not possible.",
            UserWarning,
        )
    else:
        GPU_AWARE_MPI = True



def __process_key(
    arr: DNDarray,
    key: Union[Tuple[int, ...], List[int, ...]],
    return_local_indices: Optional[bool] = False,
    op: Optional[str] = None,
) -> Tuple:
    # """
    # Private method to process the key used for indexing a ``DNDarray`` so that it can be applied to the process-local data, i.e. `key` must be "torch-proof".
    # In a processed key:
    # - any ellipses or newaxis have been replaced with the appropriate number of slice objects
    # - ndarrays and DNDarrays have been converted to torch tensors
    # - the dimensionality is the same as the ``DNDarray`` it indexes
    # This function also manipulates `arr` if necessary, inserting and/or transposing dimensions as indicated by `key`. It calculates the output shape, split axis and balanced status of the indexed array.

    # Parameters
    # ----------
    # arr : DNDarray
    #     The ``DNDarray`` to be indexed
    # key : int, Tuple[int, ...], List[int, ...]
    #     The key used for indexing
    # return_local_indices : bool, optional
    #     Whether to return the process-local indices of the key in the split dimension. This is only possible when the indexing key in the split dimension is ordered e.g. `split_key_is_ordered == 1`. Default: False
    # op : str, optional
    #     The indexing operation that the key is being processed for. Get be "get" for `__getitem__` or "set" for `__setitem__`. Default: "get".

    # Returns
    # -------
    # arr : DNDarray
    #     The ``DNDarray`` to be indexed. Its dimensions might have been modified if advanced, dimensional, broadcasted indexing is used.
    # key : Union(Tuple[Any, ...], DNDarray, np.ndarray, torch.Tensor, slice, int, List[int, ...])
    #     The processed key ready for indexing ``arr``. Its dimensions match the (potentially modified) dimensions of ``arr``.
    #     Note: the key indices along the split axis are LOCAL indices, i.e. refer to the process-local data, if ordered indexing is used. Otherwise, they are GLOBAL indices, referring to the global memory-distributed DNDarray. Communication to extract the non-ordered elements of the input ``DNDarray`` is handled by the ``__getitem__`` function.
    # output_shape : Tuple[int, ...]
    #     The shape of the output ``DNDarray``
    # new_split : int
    #     The new split axis
    # split_key_is_ordered : int
    #     Whether the split key is sorted or ordered. Can be 1: ascending, 0: not ordered, -1: descending order.
    # out_is_balanced : bool
    #     Whether the output ``DNDarray`` is balanced
    # root : int
    #     The root process for the ``MPI.Bcast`` call when single-element indexing along the split axis is used
    # backwards_transpose_axes : Tuple[int, ...]
    #     The axes to transpose the input ``DNDarray`` back to its original shape if it has been transposed for advanced indexing
    # """
    output_shape = list(arr.gshape)
    split_bookkeeping = [None] * arr.ndim
    new_split = arr.split
    arr_is_distributed = False
    if arr.split is not None:
        split_bookkeeping[arr.split] = "split"
        if arr.is_distributed():
            counts, displs = arr.counts_displs()
            arr_is_distributed = True

    advanced_indexing = False
    split_key_is_ordered = 1
    key_is_mask_like = False
    out_is_balanced = False
    root = None
    backwards_transpose_axes = tuple(range(arr.ndim))

    if isinstance(key, list):
        try:
            key = torch.tensor(key, device=arr.larray.device)
        except RuntimeError:
            raise IndexError("Invalid indices: expected a list of integers, got {}".format(key))
    if isinstance(key, (DNDarray, torch.Tensor, np.ndarray)):
        if key.dtype in (bool, uint8, torch.bool, torch.uint8, np.bool_, np.uint8):
            # boolean indexing: shape must be consistent with arr.shape
            key_ndim = key.ndim
            if not tuple(key.shape) == arr.shape[:key_ndim]:
                raise IndexError(
                    "Boolean index of shape {} does not match indexed array of shape {}".format(
                        tuple(key.shape), arr.shape
                    )
                )
            # extract non-zero elements
            try:
                # key is torch tensor
                key = key.nonzero(as_tuple=True)
            except TypeError:
                # key is np.ndarray or DNDarray
                key = key.nonzero()
            key_is_mask_like = True
        else:
            # advanced indexing on first dimension: first dim will expand to shape of key
            output_shape = tuple(list(key.shape) + output_shape[1:])
            # adjust split axis accordingly
            if arr_is_distributed:
                if arr.split != 0:
                    # split axis is not affected
                    split_bookkeeping = [None] * key.ndim + split_bookkeeping[1:]
                    new_split = (
                        split_bookkeeping.index("split")
                        if "split" in split_bookkeeping
                        else None
                    )
                    out_is_balanced = arr.balanced
                else:
                    # split axis is affected
                    if key.ndim > 1:
                        try:
                            key_numel = key.numel()
                        except AttributeError:
                            key_numel = key.size
                        if key_numel == arr.shape[0]:
                            new_split = tuple(key.shape).index(arr.shape[0])
                        else:
                            new_split = key.ndim - 1
                        try:
                            key_split = key[new_split].larray
                            sorted, _ = key_split.sort(stable=True)
                        except AttributeError:
                            key_split = key[new_split]
                            sorted = key_split.sort()
                    else:
                        new_split = 0
                        # assess if key is sorted along split axis
                        try:
                            # DNDarray key
                            sorted, _ = torch.sort(key.larray, stable=True)
                            split_key_is_ordered = torch.tensor(
                                (key.larray == sorted).all(),
                                dtype=torch.uint8,
                                device=key.larray.device,
                            )
                            if key.split is not None:
                                out_is_balanced = key.balanced
                                split_key_is_ordered = (
                                    factories.array(
                                        [split_key_is_ordered],
                                        is_split=0,
                                        device=arr.device,
                                        copy=False,
                                    )
                                    .all()
                                    .astype(types.canonical_heat_types.uint8)
                                    .item()
                                )
                            else:
                                split_key_is_ordered = split_key_is_ordered.item()
                            key = key.larray
                        except AttributeError:
                            # torch or ndarray key
                            try:
                                sorted, _ = torch.sort(key, stable=True)
                            except TypeError:
                                # ndarray key
                                sorted = torch.tensor(np.sort(key), device=arr.larray.device)
                            split_key_is_ordered = (key == sorted).all().item()
                            if not split_key_is_ordered:
                                # prepare for distributed non-ordered indexing: distribute torch/numpy key
                                key = factories.array(key, split=0, device=arr.device).larray
                                out_is_balanced = True
                        if split_key_is_ordered:
                            # extract local key
                            cond1 = key >= displs[arr.comm.rank]
                            cond2 = key < displs[arr.comm.rank] + counts[arr.comm.rank]
                            key = key[cond1 & cond2]
                            if return_local_indices:
                                key -= displs[arr.comm.rank]
                            out_is_balanced = False
            else:
                try:
                    out_is_balanced = key.balanced
                    new_split = key.split
                    key = key.larray
                except AttributeError:
                    # torch or numpy key, non-distributed indexed array
                    out_is_balanced = True
                    new_split = None
            return (
                arr,
                key,
                output_shape,
                new_split,
                split_key_is_ordered,
                key_is_mask_like,
                out_is_balanced,
                root,
                backwards_transpose_axes,
            )

    key = list(key) if isinstance(key, Iterable) else [key]

    # check for ellipsis, newaxis. NB: (np.newaxis is None)==True
    add_dims = sum(k is None for k in key)
    ellipsis = sum(isinstance(k, type(...)) for k in key)
    if ellipsis > 1:
        raise ValueError("indexing key can only contain 1 Ellipsis (...)")
    if ellipsis:
        # key contains exactly 1 ellipsis
        # replace with explicit `slice(None)` for affected dimensions
        # output_shape, split_bookkeeping not affected
        expand_key = [slice(None)] * (arr.ndim + add_dims)
        ellipsis_index = key.index(...)
        ellipsis_dims = arr.ndim - (len(key) - ellipsis - add_dims)
        expand_key[:ellipsis_index] = key[:ellipsis_index]
        expand_key[ellipsis_index + ellipsis_dims :] = key[ellipsis_index + 1 :]
        key = expand_key
    while add_dims > 0:
        # expand array dims: output_shape, split_bookkeeping to reflect newaxis
        # replace newaxis with slice(None) in key
        for i, k in reversed(list(enumerate(key))):
            if k is None:
                key[i] = slice(None)
                arr = arr.expand_dims(i - add_dims + 1)
                output_shape = (
                    output_shape[: i - add_dims + 1] + [1] + output_shape[i - add_dims + 1 :]
                )
                split_bookkeeping = (
                    split_bookkeeping[: i - add_dims + 1]
                    + [None]
                    + split_bookkeeping[i - add_dims + 1 :]
                )
                add_dims -= 1

    # recalculate new_split, transpose_axes after dimensions manipulation
    new_split = split_bookkeeping.index("split") if "split" in split_bookkeeping else None
    transpose_axes, backwards_transpose_axes = tuple(range(arr.ndim)), tuple(range(arr.ndim))
    # check for advanced indexing and slices
    advanced_indexing_dims = []
    advanced_indexing_shapes = []
    lose_dims = 0
    for i, k in enumerate(key):
        if np.isscalar(k) or getattr(k, "ndim", 1) == 0:
            # single-element indexing along axis i
            try:
                output_shape[i], split_bookkeeping[i] = None, None
            except IndexError:
                raise IndexError(
                    f"Too many indices for DNDarray: DNDarray is {arr.ndim}-dimensional, but {len(key)} dimensions were indexed"
                )
            lose_dims += 1
            if i == arr.split:
                key[i], root = arr.__process_scalar_key(
                    k, indexed_axis=i, return_local_indices=return_local_indices
                )
            else:
                key[i], _ = arr.__process_scalar_key(
                    k, indexed_axis=i, return_local_indices=False
                )
        elif isinstance(k, Iterable) or isinstance(k, DNDarray):
            advanced_indexing = True
            advanced_indexing_dims.append(i)
            # work with DNDarrays to assess distribution
            # torch tensors will be extracted in the advanced indexing section below
            k = factories.array(k, device=arr.device, comm=arr.comm, copy=None)
            advanced_indexing_shapes.append(k.gshape)
            if arr_is_distributed and i == arr.split:
                if (
                    not k.is_distributed()
                    and k.ndim == 1
                    and (k.larray == torch.sort(k.larray, stable=True)[0]).all()
                ):
                    split_key_is_ordered = 1
                    out_is_balanced = None
                else:
                    split_key_is_ordered = 0
                    # redistribute key along last axis to match split axis of indexed array
                    k = k.resplit(-1)
                    out_is_balanced = True
            key[i] = k

        elif isinstance(k, slice) and k != slice(None):
            start, stop, step = k.start, k.stop, k.step
            if start is None:
                start = 0
            elif start < 0:
                start += arr.gshape[i]
            if stop is None:
                stop = arr.gshape[i]
            elif stop < 0:
                stop += arr.gshape[i]
            if step is None:
                step = 1
            if step < 0 and start > stop:
                # PyTorch doesn't support negative step as of 1.13
                # Lazy solution, potentially large memory footprint
                # TODO: implement ht.fromiter (implemented in ASSET_ht)
                key[i] = torch.tensor(list(range(start, stop, step)), device=arr.larray.device)
                output_shape[i] = len(key[i])
                split_key_is_ordered = -1
                if arr_is_distributed and new_split == i:
                    if op == "set":
                        # setitem: flip key and keep process-local indices
                        key[i] = key[i].flip(0)
                        cond1 = key[i] >= displs[arr.comm.rank]
                        cond2 = key[i] < displs[arr.comm.rank] + counts[arr.comm.rank]
                        key[i] = key[i][cond1 & cond2]
                        if return_local_indices:
                            key[i] -= displs[arr.comm.rank]
                    else:
                        # getitem: distribute key and proceed with non-ordered indexing
                        key[i] = factories.array(
                            key[i], split=0, device=arr.device, copy=False
                        ).larray
                        out_is_balanced = True
            elif step > 0 and start < stop:
                output_shape[i] = int(torch.tensor((stop - start) / step).ceil().item())
                if arr_is_distributed and new_split == i:
                    split_key_is_ordered = 1
                    out_is_balanced = False
                    local_arr_end = displs[arr.comm.rank] + counts[arr.comm.rank]
                    if stop > displs[arr.comm.rank] and start < local_arr_end:
                        index_in_cycle = (displs[arr.comm.rank] - start) % step
                        if start >= displs[arr.comm.rank]:
                            # slice begins on current rank
                            local_start = start - displs[arr.comm.rank]
                        else:
                            local_start = 0 if index_in_cycle == 0 else step - index_in_cycle
                        if stop <= local_arr_end:
                            # slice ends on current rank
                            local_stop = stop - displs[arr.comm.rank]
                        else:
                            local_stop = local_arr_end
                        key[i] = slice(local_start, local_stop, step)
                    else:
                        key[i] = slice(0, 0)
            elif step == 0:
                raise ValueError("Slice step cannot be zero")
            else:
                key[i] = slice(0, 0)
                output_shape[i] = 0

    if advanced_indexing:
        # adv indexing key elements are DNDarrays: extract torch tensors
        # options: 1. key is mask-like (covers boolean mask as well), 2. adv indexing along split axis, 3. everything else
        # 1. define key as mask-like if each element of key is a DNDarray, and all elements of key are of the same shape, and the advanced-indexing dimensions are consecutive
        key_is_mask_like = (
            all(isinstance(k, DNDarray) for k in key)
            and len(set(k.shape for k in key)) == 1
            and torch.tensor(advanced_indexing_dims).diff().eq(1).all()
        )
        # if split axis is affected by advanced indexing, keep track of non-split dimensions for later
        if arr.is_distributed() and arr.split in advanced_indexing_dims:
            non_split_dims = list(advanced_indexing_dims).copy()
            if arr.split is not None:
                non_split_dims.remove(arr.split)
        # 1. key is mask-like
        if key_is_mask_like:
            key = list(key)
            key_splits = [k.split for k in key]
            if arr.split is not None:
                if not key_splits.count(key_splits[arr.split]) == len(key_splits):
                    if (
                        key_splits[arr.split] is not None
                        and key_splits.count(None) == len(key_splits) - 1
                    ):
                        for i in non_split_dims:
                            key[i] = factories.array(
                                key[i],
                                split=key_splits[arr.split],
                                device=arr.device,
                                comm=arr.comm,
                                copy=None,
                            )
                    else:
                        raise IndexError(
                            f"Indexing arrays must be distributed along the same dimension, got splits {key_splits}."
                        )
                else:
                    # all key_splits must be the same, otherwise raise IndexError
                    if not key_splits.count(key_splits[0]) == len(key_splits):
                        raise IndexError(
                            f"Indexing arrays must be distributed along the same dimension, got splits {key_splits}."
                        )
            # all key elements are now DNDarrays of the same shape, same split axis
        # 2. advanced indexing along split axis
        if arr.is_distributed() and arr.split in advanced_indexing_dims:
            if split_key_is_ordered == 1:
                # extract torch tensors, keep process-local indices only
                k = key[arr.split].larray
                cond1 = k >= displs[arr.comm.rank]
                cond2 = k < displs[arr.comm.rank] + counts[arr.comm.rank]
                k = k[cond1 & cond2]
                if return_local_indices:
                    k -= displs[arr.comm.rank]
                key[arr.split] = k
                for i in non_split_dims:
                    if key_is_mask_like:
                        # select the same elements along non-split dimensions
                        key[i] = key[i].larray[cond1 & cond2]
                    else:
                        key[i] = key[i].larray
            elif split_key_is_ordered == 0:
                # extract torch tensors, any other communication + mask-like case are handled in __getitem__ or __setitem__
                for i in advanced_indexing_dims:
                    key[i] = key[i].larray
            # split_key_is_ordered == -1 not treated here as it is slicing, not advanced indexing
        else:
            # advanced indexing does not affect split axis, return torch tensors
            for i in advanced_indexing_dims:
                key[i] = key[i].larray
        # all adv indexing keys are now torch tensors

        # shapes of adv indexing arrays must be broadcastable
        try:
            broadcasted_shape = torch.broadcast_shapes(*advanced_indexing_shapes)
        except RuntimeError:
            raise IndexError(
                "Shape mismatch: indexing arrays could not be broadcast together with shapes: {}".format(
                    advanced_indexing_shapes
                )
            )
        add_dims = len(broadcasted_shape) - len(advanced_indexing_dims)
        if (
            len(advanced_indexing_dims) == 1
            or list(range(advanced_indexing_dims[0], advanced_indexing_dims[-1] + 1))
            == advanced_indexing_dims
        ):
            # dimensions affected by advanced indexing are consecutive:
            output_shape[
                advanced_indexing_dims[0] : advanced_indexing_dims[0]
                + len(advanced_indexing_dims)
            ] = broadcasted_shape
            if key_is_mask_like:
                # advanced indexing dimensions will be collapsed into one dimension
                if (
                    "split" in split_bookkeeping
                    and split_bookkeeping.index("split") in advanced_indexing_dims
                ):
                    split_bookkeeping[
                        advanced_indexing_dims[0] : advanced_indexing_dims[0]
                        + len(advanced_indexing_dims)
                    ] = ["split"]
                else:
                    split_bookkeeping[
                        advanced_indexing_dims[0] : advanced_indexing_dims[0]
                        + len(advanced_indexing_dims)
                    ] = [None]
            else:
                split_bookkeeping = (
                    split_bookkeeping[: advanced_indexing_dims[0]]
                    + [None] * add_dims
                    + split_bookkeeping[advanced_indexing_dims[0] :]
                )
        else:
            # advanced-indexing dimensions are not consecutive:
            # transpose array to make the advanced-indexing dimensions consecutive as the first dimensions
            non_adv_ind_dims = list(
                i for i in range(arr.ndim) if i not in advanced_indexing_dims
            )
            # keep track of transpose axes order, to be able to transpose back later
            transpose_axes = tuple(advanced_indexing_dims + non_adv_ind_dims)
            arr = arr.transpose(transpose_axes)
            backwards_transpose_axes = tuple(
                torch.tensor(transpose_axes, device=arr.larray.device)
                .argsort(stable=True)
                .tolist()
            )
            # output shape and split bookkeeping
            output_shape = list(output_shape[i] for i in transpose_axes)
            output_shape[: len(advanced_indexing_dims)] = broadcasted_shape
            split_bookkeeping = list(split_bookkeeping[i] for i in transpose_axes)
            split_bookkeeping = [None] * add_dims + split_bookkeeping
            # modify key to match the new dimension order
            key = [key[i] for i in advanced_indexing_dims] + [key[i] for i in non_adv_ind_dims]
            # update advanced-indexing dims
            advanced_indexing_dims = list(range(len(advanced_indexing_dims)))

    # expand key to match the number of dimensions of the DNDarray
    if arr.ndim > len(key):
        key += [slice(None)] * (arr.ndim - len(key))

    key = tuple(key)
    for i in range(output_shape.count(None)):
        lost_dim = output_shape.index(None)
        output_shape.remove(None)
        split_bookkeeping = split_bookkeeping[:lost_dim] + split_bookkeeping[lost_dim + 1 :]
    output_shape = tuple(output_shape)
    new_split = split_bookkeeping.index("split") if "split" in split_bookkeeping else None
    return (
        arr,
        key,
        output_shape,
        new_split,
        split_key_is_ordered,
        key_is_mask_like,
        out_is_balanced,
        root,
        backwards_transpose_axes,
    )


"""
Collection of solvers for systems of linear equations.
"""

import heat as ht
from ..dndarray import DNDarray
from ..sanitation import sanitize_out
from typing import List, Dict, Any, TypeVar, Union, Tuple, Optional
from .. import factories

import torch

__all__ = ["cg", "lanczos", "solve", "solve_triangular"]


def cg(A: DNDarray, b: DNDarray, x0: DNDarray, out: Optional[DNDarray] = None) -> DNDarray:
    """
    Conjugate gradients method for solving a system of linear equations :math: `Ax = b`

    Parameters
    ----------
    A : DNDarray
        2D symmetric, positive definite Matrix
    b : DNDarray
        1D vector
    x0 : DNDarray
        Arbitrary 1D starting vector
    out : DNDarray, optional
        Output Vector
    """
    if not isinstance(A, DNDarray) or not isinstance(b, DNDarray) or not isinstance(x0, DNDarray):
        raise TypeError(
            f"A, b and x0 need to be of type ht.DNDarray, but were {type(A)}, {type(b)}, {type(x0)}"
        )

    if A.ndim != 2:
        raise RuntimeError("A needs to be a 2D matrix")
    if b.ndim != 1:
        raise RuntimeError("b needs to be a 1D vector")
    if x0.ndim != 1:
        raise RuntimeError("c needs to be a 1D vector")

    r = b - ht.matmul(A, x0)
    p = r
    rsold = ht.matmul(r, r)
    x = x0

    for i in range(len(b)):
        Ap = ht.matmul(A, p)
        alpha = rsold / ht.matmul(p, Ap)
        x = x + alpha * p
        r = r - alpha * Ap
        rsnew = ht.matmul(r, r)
        if ht.sqrt(rsnew).item() < 1e-10:
            print(f"Residual reaches tolerance in it = {i}")
            if out is not None:
                out = x
                return out
            return x
        p = r + ((rsnew / rsold) * p)
        rsold = rsnew

    if out is not None:
        out = x
        return out
    return x


def lanczos(
    A: DNDarray,
    m: int,
    v0: Optional[DNDarray] = None,
    V_out: Optional[DNDarray] = None,
    T_out: Optional[DNDarray] = None,
) -> Tuple[DNDarray, DNDarray]:
    r"""
    The Lanczos algorithm is an iterative approximation of the solution to the eigenvalue problem, as an adaptation of
    power methods to find the m "most useful" (tending towards extreme highest/lowest) eigenvalues and eigenvectors of
    an :math:`n \times n` Hermitian matrix, where often :math:`m<<n`.
    It returns two matrices :math:`V` and :math:`T`, where:

        - :math:`V` is a Matrix of size :math:`n\times m`, with orthonormal columns, that span the Krylow subspace \n
        - :math:`T` is a Tridiagonal matrix of size :math:`m\times m`, with coefficients :math:`\alpha_1,..., \alpha_n`
          on the diagonal and coefficients :math:`\beta_1,...,\beta_{n-1}` on the side-diagonals\n

    Parameters
    ----------
    A : DNDarray
        2D Hermitian (if complex) or symmetric positive-definite matrix.
        Only distribution along axis 0 is supported, i.e. `A.split` must be `0` or `None`.
    m : int
        Number of Lanczos iterations
    v0 : DNDarray, optional
        1D starting vector of Euclidean norm 1. If not provided, a random vector will be used to start the algorithm
    V_out : DNDarray, optional
        Output Matrix for the Krylow vectors, Shape = (n, m), dtype=A.dtype, must be initialized to zero
    T_out : DNDarray, optional
        Output Matrix for the Tridiagonal matrix, Shape = (m, m), must be initialized to zero
    """
    if not isinstance(A, DNDarray):
        raise TypeError(f"A needs to be of type ht.dndarray, but was {type(A)}")
    if A.ndim != 2:
        raise RuntimeError("A needs to be a 2D matrix")
    if A.dtype is ht.int32 or A.dtype is ht.int64:
        raise TypeError(f"A can be float or complex, got {A.dtype}")
    if not isinstance(m, (int, float)):
        raise TypeError(f"m must be int, got {type(m)}")

    n, column = A.shape
    if n != column:
        raise TypeError("Input Matrix A needs to be symmetric positive-definite.")

    # output data types: T is always Real
    A_is_complex = A.dtype is ht.complex128 or A.dtype is ht.complex64
    T_dtype = A.real.dtype

    # initialize or sanitize output buffers
    if T_out is not None:
        sanitize_out(
            T_out,
            output_shape=(m, m),
            output_split=None,
            output_device=A.device,
            output_comm=A.comm,
        )
        T = T_out
    else:
        T = ht.zeros((m, m), dtype=T_dtype, device=A.device, comm=A.comm)
    if A.split == 0:
        if V_out is not None:
            sanitize_out(
                V_out,
                output_shape=(n, m),
                output_split=0,
                output_device=A.device,
                output_comm=A.comm,
            )
            V = V_out
        else:
            # This is done for better memory access in the reorthogonalization Gram-Schmidt algorithm
            V = ht.zeros((n, m), split=0, dtype=A.dtype, device=A.device, comm=A.comm)
    else:
        if A.split == 1:
            raise NotImplementedError("Distribution along axis 1 not implemented yet.")
        if V_out is not None:
            sanitize_out(
                V_out,
                output_shape=(n, m),
                output_split=None,
                output_device=A.device,
                output_comm=A.comm,
            )
            V = V_out
        else:
            V = ht.zeros((n, m), split=None, dtype=A.dtype, device=A.device, comm=A.comm)

    if A_is_complex:
        if v0 is None:
            vr = (
                ht.random.rand(n, split=V.split, dtype=T_dtype, device=V.device, comm=V.comm)
                + ht.random.rand(n, split=V.split, dtype=T_dtype, device=V.device, comm=V.comm) * 1j
            )
            v0 = vr / ht.norm(vr)
        elif v0.split != V.split:
            v0.resplit_(axis=V.split)
        # # 0th iteration
        # # vector v0 has Euclidean norm = 1
        w = ht.matmul(A, v0)
        alpha = ht.dot(ht.conj(w).T, v0)
        w = w - alpha * v0
        T[0, 0] = alpha.real
        V[:, 0] = v0
        for i in range(1, int(m)):
            beta = ht.norm(w)
            if ht.abs(beta) < 1e-10:
                # print("Lanczos breakdown in iteration {}".format(i))
                # Lanczos Breakdown, pick a random vector to continue
                vr = (
                    ht.random.rand(n, split=V.split, dtype=T_dtype, device=V.device, comm=V.comm)
                    + ht.random.rand(n, split=V.split, dtype=T_dtype, device=V.device, comm=V.comm)
                    * 1j
                )
                # orthogonalize v_r with respect to all vectors v[i]
                for j in range(i):
                    vi_loc = V._DNDarray__array[:, j]
                    a = torch.dot(vr.larray, torch.conj(vi_loc))
                    b = torch.dot(vi_loc, torch.conj(vi_loc))
                    A.comm.Allreduce(ht.communication.MPI.IN_PLACE, a, ht.communication.MPI.SUM)
                    A.comm.Allreduce(ht.communication.MPI.IN_PLACE, b, ht.communication.MPI.SUM)
                    vr._DNDarray__array = vr._DNDarray__array - a / b * vi_loc
                # normalize v_r to Euclidean norm 1 and set as ith vector v
                vi = vr / ht.norm(vr)
            else:
                vr = w

                # Reorthogonalization
                for j in range(i):
                    vi_loc = V.larray[:, j]
                    a = torch.dot(vr._DNDarray__array, torch.conj(vi_loc))
                    b = torch.dot(vi_loc, torch.conj(vi_loc))
                    A.comm.Allreduce(ht.communication.MPI.IN_PLACE, a, ht.communication.MPI.SUM)
                    A.comm.Allreduce(ht.communication.MPI.IN_PLACE, b, ht.communication.MPI.SUM)
                    vr._DNDarray__array = vr._DNDarray__array - a / b * vi_loc

                vi = vr / ht.norm(vr)

            w = ht.matmul(A, vi)
            alpha = ht.dot(ht.conj(w).T, vi)

            w = w - alpha * vi - beta * V[:, i - 1]

            T[i - 1, i] = beta.real
            T[i, i - 1] = beta.real
            T[i, i] = alpha.real
            V[:, i] = vi
    else:
        if v0 is None:
            vr = ht.random.rand(n, split=V.split, dtype=T_dtype, device=V.device, comm=V.comm)
            v0 = vr / ht.norm(vr)
        elif v0.split != V.split:
            v0.resplit_(axis=V.split)
        # # 0th iteration
        # # vector v0 has Euclidean norm = 1
        w = ht.matmul(A, v0)
        alpha = ht.dot(w, v0)
        w = w - alpha * v0
        T[0, 0] = alpha
        V[:, 0] = v0
        for i in range(1, int(m)):
            beta = ht.norm(w)
            if ht.abs(beta) < 1e-10:
                # print("Lanczos breakdown in iteration {}".format(i))
                # Lanczos Breakdown, pick a random vector to continue
                vr = ht.random.rand(n, split=V.split, dtype=T_dtype, device=V.device, comm=V.comm)
                # orthogonalize v_r with respect to all vectors v[i]
                for j in range(i):
                    vi_loc = V._DNDarray__array[:, j]
                    a = torch.dot(vr.larray, vi_loc)
                    b = torch.dot(vi_loc, vi_loc)
                    A.comm.Allreduce(ht.communication.MPI.IN_PLACE, a, ht.communication.MPI.SUM)
                    A.comm.Allreduce(ht.communication.MPI.IN_PLACE, b, ht.communication.MPI.SUM)
                    vr._DNDarray__array = vr._DNDarray__array - a / b * vi_loc
                # normalize v_r to Euclidean norm 1 and set as ith vector v
                vi = vr / ht.norm(vr)
            else:
                vr = w

                # Reorthogonalization
                for j in range(i):
                    vi_loc = V.larray[:, j]
                    a = torch.dot(vr._DNDarray__array, vi_loc)
                    b = torch.dot(vi_loc, vi_loc)
                    A.comm.Allreduce(ht.communication.MPI.IN_PLACE, a, ht.communication.MPI.SUM)
                    A.comm.Allreduce(ht.communication.MPI.IN_PLACE, b, ht.communication.MPI.SUM)
                    vr._DNDarray__array = vr._DNDarray__array - a / b * vi_loc

                vi = vr / ht.norm(vr)

            w = ht.matmul(A, vi)
            alpha = ht.dot(w, vi)

            w = w - alpha * vi - beta * V[:, i - 1]

            T[i - 1, i] = beta
            T[i, i - 1] = beta
            T[i, i] = alpha
            V[:, i] = vi

    if V.split is not None:
        V.resplit_(axis=None)

    return V, T


def solve(A: DNDarray, b: DNDarray, out: Optional[DNDarray] = None) -> DNDarray:
    r"""
    Computes the solution of a square system of linear equations with a unique solution.

    Letting :math:`\mathbb{K}` be :math:`\mathbb{R}` or :math:`\mathbb{C}`,
    this function computes the solution :math:`X \in \mathbb{K}^{n \times k}` of the **linear system** associated to
    :math:`A \in \mathbb{K}^{n \times n}, B \in \mathbb{K}^{n \times k}`, which is defined as

    .. math:: AX = B

    Supports inputs of integer, float, double, cfloat and cdouble dtypes.
    Also supports batches of matrices, and if the inputs are batches of matrices then
    the output has the same batch dimensions.

    Letting `*` be zero or more batch dimensions,

    - If :attr:`A` has shape `(*, n, n)` and :attr:`B` has shape `(*, n)` (a batch of vectors) or shape
      `(*, n, k)` (a batch of matrices or "multiple right-hand sides"), this function returns `X` of shape
      `(*, n)` or `(*, n, k)` respectively.
    - Otherwise, if :attr:`A` has shape `(*, n, n)` and  :attr:`B` has shape `(n,)`  or `(n, k)`, :attr:`B`
      is broadcast to have shape `(*, n)` or `(*, n, k)` respectively.
      This function then returns the solution of the resulting batch of systems of linear equations.

    .. note::
        A and b may only be distributed in the batch dimensions. If both are split, they must be split along matching batch axes.

    .. seealso::

            :func:`torch.linalg.solve` is called under the hood on the local data. This docstring is also heavily inspired by the docstring of this function.


    Parameters
    ----------
    A : DNDarray
        Matrix to be inverted of shape `(*, n, n)` where `*` is zero or more batch dimensions
    b : DNDarray
        Right-hand side of shape `(*, n)` or  `(*, n, k)` or `(n,)` or `(n, k)`
    out : DNDarray, optional
        Output Vector

    Examples::

        >>> A = ht.random.randn(3, 3)
        >>> b = ht.random.randn(3)
        >>> x = ht.linalg.solve(A, b)
        >>> ht.allclose(A @ x, b, atol=1e-5)
        True
        >>> A = ht.random.randn(2, 3, 3, split=0)
        >>> B = ht.random.randn(2, 3, 4, split=0)
        >>> X = ht.linalg.solve(A, B)
        >>> X.shape
        (2, 3, 4)
        >>> ht.allclose(A @ X, B, atol=1e-5)
        True
        >>> A = ht.random.randn(2, 3, 3, split=None)
        >>> B = ht.random.randn(2, 3, 4, split=2)
        >>> X = ht.linalg.solve(A, B)
        >>> X.split
        2
        >>> ht.allclose(A @ X, B, atol=1e-5)
        True

        >>> A = ht.random.randn(2, 3, 3, split=0)
        >>> b = ht.random.randn(3, 1)
        >>> x = ht.linalg.solve(A, b) # b is broadcast to size (2, 3, 1)
        >>> x.shape
        (2, 3, 1)
        >>> x.split
        0
        >>> ht.allclose((A @ x).resplit_(None), b, atol=1e-5)
        True
    """
    ht.sanitize_in(A)
    ht.sanitize_in(b)

    # torch doesn't support integer, so we cast to float here if needed
    if ht.issubdtype(A.dtype, ht.integer):
        A = A.astype(ht.promote_types(A.dtype, ht.float), copy=True)
    if ht.issubdtype(b.dtype, ht.integer):
        b = b.astype(ht.promote_types(b.dtype, ht.float), copy=True)

    # figure out which is the non-batched axis in b
    if b.shape[-1] == A.shape[-1]:
        b_non_batched_axis = b.ndim - 1
    elif b.ndim == 1:
        raise ValueError(f"b has incorrect shape of {b.shape} for A of shape {A.shape}")
    elif b.shape[-2] == A.shape[-1]:
        b_non_batched_axis = b.ndim - 2
    else:
        raise ValueError(f"b has incorrect shape of {b.shape} for A of shape {A.shape}")

    # raise error if b is distributed in disallowed way
    if b.is_distributed() and b.split == b_non_batched_axis:
        raise NotImplementedError(
            f"b of shape {b.shape} with A of shape {A.shape} is split in {b.split} which is the non-batched axis, but it can only be distributed in batched axes in this implementation. If you require this feature, please open an issue on GitHub."
        )

    # raise errors if A is distributed in disallowed way
    if A.is_distributed():
        if A.split > A.ndim - 3:
            raise NotImplementedError(
                f"A of dimension {A.ndim} is split in {A.split} but must not be distributed in the (non-batched) last two axes in this implementation. If you require this feature, please open an issue on GitHub."
            )
        elif A.split != b.split and b.is_distributed():
            raise ValueError(f"Split of A and b must match, but got {A.split} and {b.split}")

    # figure out what the output vector looks like
    out_initalization = {"dtype": ht.types.promote_types(A.dtype, b.dtype), "device": b.device}
    if b.shape[:b_non_batched_axis] == A.shape[:-2]:  # no need for expansion
        out_initalization["shape"] = b.shape
        out_initalization["split"] = b.split
        out_initalization["comm"] = b.comm
    elif b_non_batched_axis == 0 and A.ndim > 2:  # b needs expanding
        out_initalization["shape"] = A.shape[:-2] + b.shape
        if A.split is None:
            out_initalization["split"] = b.split
            out_initalization["comm"] = b.comm
        else:
            out_initalization["split"] = A.split
            out_initalization["comm"] = A.comm
    else:
        raise ValueError(
            f"Don't know how to batch solve with A of shape {A.shape} and split {A.split} and b of shape {b.shape} and split {b.split}"
        )

    # set up output vector
    if out is None:
        out = factories.empty(**out_initalization)

    ht.sanitize_out(
        out,
        output_shape=out_initalization["shape"],
        output_split=out_initalization["split"],
        output_device=out_initalization["device"],
        output_comm=out_initalization["comm"],
    )

    # if out is integer, we may need to cast to float here
    if out.dtype != out_initalization["dtype"]:
        out = out.astype(out_initalization["dtype"], copy=False)

    # do the actual solving of local matrices in torch
    try:
        torch.linalg.solve(A.larray, b.larray, left=True, out=out.larray)
        error_msg = ""
    except torch._C._LinAlgError as e:
        error_msg = str(e)

    global_error_msg = [error_msg for error_msg in ht.comm.allgather(error_msg) if error_msg != ""]
    if len(global_error_msg):
        raise RuntimeError(*global_error_msg)

    return out


def solve_triangular(A: DNDarray, b: DNDarray) -> DNDarray:
    """
    Solver for (possibly batched) upper triangular systems of linear equations: it returns `x` in `Ax = b`, where `A` is a (possibly batched) upper triangular matrix and
    `b` a (possibly batched) vector or matrix of suitable shape, both provided as input to the function.
    The implementation builts on the corresponding solver in PyTorch and implements an memory-distributed, MPI-parallel block-wise version thereof.

    Parameters
    ----------
    A : DNDarray
        An upper triangular invertible square (n x n) matrix or a batch thereof,  i.e. a ``DNDarray`` of shape `(..., n, n)`.
    b : DNDarray
        a (possibly batched) n x k matrix, i.e. an DNDarray of shape (..., n, k), where the batch-dimensions denoted by ... need to coincide with those of A.
        (Batched) Vectors have to be provided as ... x n x 1 matrices and the split dimension of b must the second last dimension if not None.
    Note
    ---------
    Since such a check might be computationally expensive, we do not check whether A is indeed upper triangular.
    If you require such a check, please open an issue on our GitHub page and request this feature.
    """
    if not isinstance(A, DNDarray) or not isinstance(b, DNDarray):
        raise TypeError(f"Arguments need to be of type DNDarray, got {type(A)}, {type(b)}.")
    if not A.ndim >= 2:
        raise ValueError("A needs to be a (batched) matrix.")
    if not b.ndim == A.ndim:
        raise ValueError("b needs to have the same number of (batch) dimensions as A.")
    if not A.shape[-2] == A.shape[-1]:
        raise ValueError("A needs to be a (batched) square matrix.")

    batch_dim = A.ndim - 2
    batch_shape = A.shape[:batch_dim]

    if not A.shape[:batch_dim] == b.shape[:batch_dim]:
        raise ValueError("Batch dimensions of A and b must be of the same shape.")
    if b.split == batch_dim + 1:
        raise ValueError("split=1 is not allowed for the right hand side.")
    if not b.shape[batch_dim] == A.shape[-1]:
        raise ValueError("Dimension mismatch of A and b.")

    if (
        A.split is not None and A.split < batch_dim or b.split is not None and b.split < batch_dim
    ):  # batch split
        if A.split != b.split:
            raise ValueError(
                "If a split dimension is a batch dimension, A and b must have the same split dimension. A possible solution would be a resplit of A or b to the same split dimension."
            )
    else:
        if (
            A.split is not None and b.split is not None
        ):  # both la dimensions split --> b.split = batch_dim
            # TODO remove?
            if not all(A.lshape_map[:, A.split] == b.lshape_map[:, batch_dim]):
                raise RuntimeError(
                    "The process-local arrays of A and b have different sizes along the splitted axis. This is most likely due to one of the DNDarrays being in unbalanced state. \n Consider using `A.is_balanced(force_check=True)` and `b.is_balanced(force_check=True)` to check if A and b are balanced; \n then call `A.balance_()` and/or `b.balance_()` in order to achieve equal local shapes along the split axis before applying `solve_triangular`."
                )

    comm = A.comm
    dev = A.device
    tdev = dev.torch_device

    nprocs = comm.Get_size()

    if A.split is None:  # A not split
        if b.split is None:
            x = torch.linalg.solve_triangular(A.larray, b.larray, upper=True)

            return factories.array(x, dtype=b.dtype, device=dev, comm=comm)
        else:  # A not split, b.split == -2
            b_lshapes_cum = torch.hstack(
                [
                    torch.zeros(1, dtype=torch.int64, device=tdev),
                    torch.cumsum(b.lshape_map[:, -2], 0),
                ]
            )

            btilde_loc = b.larray.clone()
            A_loc = A.larray[..., b_lshapes_cum[comm.rank] : b_lshapes_cum[comm.rank + 1]]

            x = factories.zeros_like(b, device=dev, comm=comm)

            for i in range(nprocs - 1, 0, -1):
                count = x.lshape_map[:, batch_dim].to(torch.device("cpu")).clone().numpy()
                displ = b_lshapes_cum[:-1].to(torch.device("cpu")).clone().numpy()
                count[i:] = 0  # nothing to send, as there are only zero rows
                displ[i:] = 0

                res_send = torch.empty(0)
                res_recv = torch.zeros((*batch_shape, count[comm.rank], b.shape[-1]), device=tdev)

                if comm.rank == i:
                    x.larray = torch.linalg.solve_triangular(
                        A_loc[..., b_lshapes_cum[i] : b_lshapes_cum[i + 1], :],
                        btilde_loc,
                        upper=True,
                    )
                    res_send = A_loc @ x.larray

                comm.Scatterv((res_send, count, displ), res_recv, root=i, axis=batch_dim)

                if comm.rank < i:
                    btilde_loc -= res_recv

            if comm.rank == 0:
                x.larray = torch.linalg.solve_triangular(
                    A_loc[..., : b_lshapes_cum[1], :], btilde_loc, upper=True
                )

            return x

    if A.split < batch_dim:  # batch split
        x = factories.zeros_like(b, device=dev, comm=comm, split=A.split)
        x.larray = torch.linalg.solve_triangular(A.larray, b.larray, upper=True)

        return x

    if A.split >= batch_dim:  # both splits in la dims
        A_lshapes_cum = torch.hstack(
            [
                torch.zeros(1, dtype=torch.int64, device=tdev),
                torch.cumsum(A.lshape_map[:, A.split], 0),
            ]
        )

        if b.split is None:
            btilde_loc = b.larray[
                ..., A_lshapes_cum[comm.rank] : A_lshapes_cum[comm.rank + 1], :
            ].clone()
        else:  # b is split at la dim 0
            btilde_loc = b.larray.clone()

        x = factories.zeros_like(
            b, device=dev, comm=comm, split=batch_dim
        )  # split at la dim 0 in case b is not split

        if A.split == batch_dim + 1:
            for i in range(nprocs - 1, 0, -1):
                count = x.lshape_map[:, batch_dim].to(torch.device("cpu")).clone().numpy()
                displ = A_lshapes_cum[:-1].to(torch.device("cpu")).clone().numpy()
                count[i:] = 0  # nothing to send, as there are only zero rows
                displ[i:] = 0

                res_send = torch.empty(0)
                res_recv = torch.zeros(
                    (*batch_shape, count[comm.rank], b.shape[-1]),
                    device=tdev,
                    dtype=b.dtype.torch_type(),
                )

                if comm.rank == i:
                    x.larray = torch.linalg.solve_triangular(
                        A.larray[..., A_lshapes_cum[i] : A_lshapes_cum[i + 1], :],
                        btilde_loc,
                        upper=True,
                    )
                    res_send = A.larray @ x.larray

                comm.Scatterv((res_send, count, displ), res_recv, root=i, axis=batch_dim)

                if comm.rank < i:
                    btilde_loc -= res_recv

            if comm.rank == 0:
                x.larray = torch.linalg.solve_triangular(
                    A.larray[..., : A_lshapes_cum[1], :], btilde_loc, upper=True
                )

        else:  # split dim is la dim 0
            for i in range(nprocs - 1, 0, -1):
                idims = tuple(x.lshape_map[i])
                if comm.rank == i:
                    x.larray = torch.linalg.solve_triangular(
                        A.larray[..., :, A_lshapes_cum[i] : A_lshapes_cum[i + 1]],
                        btilde_loc,
                        upper=True,
                    )
                    x_from_i = x.larray
                else:
                    x_from_i = torch.zeros(
                        idims,
                        dtype=b.dtype.torch_type(),
                        device=tdev,
                    )

                comm.Bcast(x_from_i, root=i)

                if comm.rank < i:
                    btilde_loc -= (
                        A.larray[..., :, A_lshapes_cum[i] : A_lshapes_cum[i + 1]] @ x_from_i
                    )

            if comm.rank == 0:
                x.larray = torch.linalg.solve_triangular(
                    A.larray[..., :, : A_lshapes_cum[1]], btilde_loc, upper=True
                )

        return x



"""
A Heat sub-namespace that conforms to the Python array API standard.
"""

import warnings

warnings.warn("The heat.array_api submodule is not fully implemented.", stacklevel=2)

__array_api_version__ = "2025.12"

__all__ = ["__array_api_version__"]

from ._inspection import __array_namespace_info__

from ._constants import e, inf, nan, newaxis, pi

__all__ += ["e", "inf", "nan", "newaxis", "pi"]

from ._creation_functions import (
    arange,
    asarray,
    empty,
    empty_like,
    eye,
    from_dlpack,
    full,
    full_like,
    linspace,
    meshgrid,
    ones,
    ones_like,
    tril,
    triu,
    zeros,
    zeros_like,
)

__all__ += [
    "arange",
    "asarray",
    "empty",
    "empty_like",
    "eye",
    "from_dlpack",
    "full",
    "full_like",
    "linspace",
    "meshgrid",
    "ones",
    "ones_like",
    "tril",
    "triu",
    "zeros",
    "zeros_like",
]

from ._data_type_functions import (
    astype,
    broadcast_arrays,
    broadcast_to,
    can_cast,
    finfo,
    iinfo,
    result_type,
)

__all__ += [
    "astype",
    "broadcast_arrays",
    "broadcast_to",
    "can_cast",
    "finfo",
    "iinfo",
    "result_type",
]

from heat.core.devices import cpu

__all__ += ["cpu"]

import heat.core.devices

if hasattr(heat.core.devices, "gpu"):
    from heat.core.devices import gpu

    __all__ += ["gpu"]

from ._dtypes import (
    bool,
    int8,
    int16,
    int32,
    int64,
    uint8,
    # uint16,
    # uint32,
    # uint64,
    float32,
    float64,
    complex64,
    complex128,
)

__all__ += [
    "bool",
    "int8",
    "int16",
    "int32",
    "int64",
    "uint8",
    # "uint16",
    # "uint32",
    # "uint64",
    "float32",
    "float64",
    "complex64",
    "complex128",
]

from ._elementwise_functions import (
    abs,
    acos,
    acosh,
    add,
    asin,
    asinh,
    atan,
    atan2,
    atanh,
    bitwise_and,
    bitwise_left_shift,
    bitwise_invert,
    bitwise_or,
    bitwise_right_shift,
    bitwise_xor,
    ceil,
    cos,
    cosh,
    divide,
    equal,
    exp,
    expm1,
    floor,
    floor_divide,
    greater,
    greater_equal,
    isfinite,
    isinf,
    isnan,
    less,
    less_equal,
    log,
    log1p,
    log2,
    log10,
    logaddexp,
    logical_and,
    logical_not,
    logical_or,
    logical_xor,
    multiply,
    negative,
    not_equal,
    positive,
    pow,
    remainder,
    round,
    sign,
    sin,
    sinh,
    square,
    sqrt,
    subtract,
    tan,
    tanh,
    trunc,
)

__all__ += [
    "abs",
    "acos",
    "acosh",
    "add",
    "asin",
    "asinh",
    "atan",
    "atan2",
    "atanh",
    "bitwise_and",
    "bitwise_left_shift",
    "bitwise_invert",
    "bitwise_or",
    "bitwise_right_shift",
    "bitwise_xor",
    "ceil",
    "cos",
    "cosh",
    "divide",
    "equal",
    "exp",
    "expm1",
    "floor",
    "floor_divide",
    "greater",
    "greater_equal",
    "isfinite",
    "isinf",
    "isnan",
    "less",
    "less_equal",
    "log",
    "log1p",
    "log2",
    "log10",
    "logaddexp",
    "logical_and",
    "logical_not",
    "logical_or",
    "logical_xor",
    "multiply",
    "negative",
    "not_equal",
    "positive",
    "pow",
    "remainder",
    "round",
    "sign",
    "sin",
    "sinh",
    "square",
    "sqrt",
    "subtract",
    "tan",
    "tanh",
    "trunc",
]

from . import linalg

__all__ += ["linalg"]

from .linalg import matmul, matrix_transpose, tensordot, vecdot

__all__ += ["matmul", "matrix_transpose", "tensordot", "vecdot"]

from ._manipulation_functions import (
    concat,
    expand_dims,
    flip,
    permute_dims,
    reshape,
    roll,
    squeeze,
    stack,
)

__all__ += ["concat", "expand_dims", "flip", "permute_dims", "reshape", "roll", "squeeze", "stack"]

from ._searching_functions import argmax, argmin, nonzero, where

__all__ += ["argmax", "argmin", "nonzero", "where"]

from ._set_functions import unique_inverse, unique_values

__all__ += ["unique_inverse", "unique_values"]

from ._sorting_functions import sort

__all__ += ["sort"]

from ._statistical_functions import max, mean, min, prod, std, sum, var

__all__ += ["max", "mean", "min", "prod", "std", "sum", "var"]

from ._utility_functions import all, any

__all__ += ["all", "any"]
